# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Imports
import os
from typing import Optional, Union
import xml.etree.ElementTree as ET

# Third Party Imports
import numpy as np
import numpy.typing as npt

# Local imports
from .metadata import XMLMetadata
from navigate.tools.linear_algebra import affine_rotation, affine_shear


class BigDataViewerMetadata(XMLMetadata):
    """Metadata for BigDataViewer files.

    Note
    ----
        XML spec in section 2.3 of https://arxiv.org/abs/1412.0488.

    """

    def __init__(self) -> None:
        """Initialize the BigDataViewer metadata object."""
        super().__init__()

        # Affine Transform Parameters
        #: bool: Shear the data.
        self.shear_data = False
        #: str: Dimension to shear the data.
        self.shear_dimension = "YZ"
        #: float: Angle in degrees to shear the data.
        self.shear_angle = 0
        #: npt.NDArray: Shear transform matrix.
        self.shear_transform = np.eye(3, 4)

        # Rotation Transform Parameters
        #: bool: Rotate the data.
        self.rotate_data = False
        #: float: Angle in degrees to rotate the data in X.
        self.rotate_angle_x = 0
        #: float: Angle in degrees to rotate the data in Y.
        self.rotate_angle_y = 0
        #: float: Angle in degrees to rotate the data in Z.
        self.rotate_angle_z = 0
        #: npt.NDArray: Rotation transform matrix.
        self.rotate_transform = np.eye(3, 4)

    def get_affine_parameters(self, configuration):
        """Get the affine transform parameters from the configuration file.

        Parameters
        ----------
        configuration : dict
            Configuration dictionary.
        """
        bdv_configuration = configuration["configuration"].get("BDVParameters")

        if bdv_configuration is not None:

            # Shear Parameters
            self.shear_data = bdv_configuration["shear"].get("shear_data", False)
            self.shear_dimension = (
                bdv_configuration["shear"].get("shear_dimension", "YZ").upper()
            )
            self.shear_angle = bdv_configuration["shear"].get("shear_angle", 0)

            # Rotate Parameters
            self.rotate_data = bdv_configuration["rotate"].get("rotate_data", False)
            self.rotate_angle_x = bdv_configuration["rotate"].get("X", 0)
            self.rotate_angle_y = bdv_configuration["rotate"].get("Y", 0)
            self.rotate_angle_z = bdv_configuration["rotate"].get("Z", 0)

    def bdv_xml_dict(
        self, file_name: Union[str, list, None], views: list, **kw
    ) -> dict:
        """Create a BigDataViewer XML dictionary from a list of views.

        Parameters
        ----------
        file_name : str
            The file name of the file to be written.
        views : list
            A list of dictionaries containing metadata for each view.
        **kw
            Additional keyword arguments.

        Returns
        -------
        dict
            A dictionary containing the XML metadata.

        """
        # Header
        bdv_dict = {"version": 0.2}

        # File path
        bdv_dict["BasePath"] = {"type": "relative", "text": "."}
        bdv_dict["SequenceDescription"] = {}

        ext = os.path.basename(file_name).split(".")[-1]
        if ext == "h5":
            """
            <ImageLoader format="bdv.hdf5">
                <hdf5 type="relative">dataset.h5</hdf5>
            </ImageLoader>
            """
            bdv_dict["SequenceDescription"]["ImageLoader"] = {"format": "bdv.hdf5"}
            bdv_dict["SequenceDescription"]["ImageLoader"]["hdf5"] = {
                "type": "relative",
                "text": file_name,
            }

        elif ext == "tiff" or ext == "tif":
            """
            Need to iterate through the time points, etc.
            <ImageLoader format="spimreconstruction.filelist">
                <imglib2container>ArrayImgFactory</imglib2container>
                <ZGrouped>false</ZGrouped>
                <files>
                    <FileMapping view_setup="0" timepoint="0" series="0" channel="0">
                        <file type="relative">1_CH00_000000.tif</file>
                    </FileMapping>
                    <FileMapping view_setup="1" timepoint="0" series="0" channel="0">
                        <file type="relative">1_CH01_000000.tif</file>
                    </FileMapping>
                </files>
            </ImageLoader>
            """
            pass

        elif ext == "n5":
            """
            <ImageLoader format="bdv.n5" version="1.0">
                <n5 type="relative">dataset.n5</n5>
            </ImageLoader>
            """
            bdv_dict["SequenceDescription"]["ImageLoader"] = {"format": "bdv.n5"}
            bdv_dict["SequenceDescription"]["ImageLoader"]["n5"] = {
                "type": "relative",
                "text": file_name,
            }

        # Calculate shear and rotation transforms
        self.bdv_shear_transform()
        self.bdv_rotate_transform()

        # Populate ViewSetups
        bdv_dict["SequenceDescription"]["ViewSetups"] = {}
        bdv_dict["SequenceDescription"]["ViewSetups"]["ViewSetup"] = []
        # Attributes are necessary for BigStitcher
        bdv_dict["SequenceDescription"]["ViewSetups"]["Attributes"] = [
            {
                "name": "illumination",
                "Illumination": {"id": {"text": 0}, "name": {"text": 0}},
            },
            {"name": "channel", "Channel": []},
            {"name": "tile", "Tile": []},
            {"name": "angle", "Angle": {"id": {"text": 0}, "name": {"text": 0}}},
        ]
        # The actual loop that populates ViewSetup
        view_id = 0
        for c in range(self.shape_c):
            # We also take care of the Channel attributes here
            ch = {"id": {"text": str(c)}, "name": {"text": str(c)}}
            bdv_dict["SequenceDescription"]["ViewSetups"]["Attributes"][1][
                "Channel"
            ].append(ch)
            for p in range(self.positions):
                d = {"id": {"text": view_id}, "name": {"text": view_id}}
                d["size"] = {"text": f"{self.shape_x} {self.shape_y} {self.shape_z}"}
                d["voxelSize"] = {"unit": {"text": "um"}}
                d["voxelSize"]["size"] = {"text": f"{self.dx} {self.dy} {self.dz}"}
                # These attributes are necessary for BigStitcher
                d["attributes"] = {
                    "illumination": {"text": "0"},
                    "channel": {"text": str(c)},
                    "tile": {"text": str(p)},
                    "angle": {"text": "0"},
                }
                bdv_dict["SequenceDescription"]["ViewSetups"]["ViewSetup"].append(d)
                view_id += 1
        # Finish up the Tile Attributes outside of the channels loop so we have
        # one per tile
        for p in range(self.positions):
            tile = {"id": {"text": str(p)}, "name": {"text": str(p)}}
            bdv_dict["SequenceDescription"]["ViewSetups"]["Attributes"][2][
                "Tile"
            ].append(tile)

        # Time
        bdv_dict["SequenceDescription"]["Timepoints"] = {"type": "range"}
        bdv_dict["SequenceDescription"]["Timepoints"]["first"] = {"text": 0}
        bdv_dict["SequenceDescription"]["Timepoints"]["last"] = {
            "text": self.shape_t - 1
        }

        # View registrations
        bdv_dict["ViewRegistrations"] = {"ViewRegistration": []}
        for t in range(self.shape_t):
            for p in range(self.positions):
                for c in range(self.shape_c):
                    view_id = c * self.positions + p
                    mat = np.zeros((3, 4), dtype=float)
                    for z in range(self.shape_z):
                        matrix_id = (
                            z
                            + self.shape_z * c
                            + p * self.shape_c * self.shape_z
                            + t * self.shape_c * self.shape_z * self.positions
                        )

                        # Construct centroid of volume matrix
                        # print(matrix_id, views[matrix_id])
                        try:
                            mat += (
                                self.stage_positions_to_affine_matrix(
                                    **views[matrix_id]
                                )
                                / self.shape_z
                            )
                        except IndexError:
                            # We have most likely canceled in the middle of
                            # an acquisition.
                            pass

                    view_transforms = [
                        {
                            "type": "affine",
                            "Name": "Translation to Regular Grid",
                            "affine": {
                                "text": " ".join([f"{x:.6f}" for x in mat.ravel()])
                            },
                        }
                    ]

                    if self.shear_data:
                        view_transforms.append(
                            {
                                "type": "affine",
                                "Name": "Shearing Transform",
                                "affine": {
                                    "text": " ".join(
                                        [
                                            f"{x:.6f}"
                                            for x in self.shear_transform.ravel()
                                        ]
                                    )
                                },
                            }
                        )

                    if self.rotate_data:
                        view_transforms.append(
                            {
                                "type": "affine",
                                "Name": "Rotation Transform",
                                "affine": {
                                    "text": " ".join(
                                        [
                                            f"{x:.6f}"
                                            for x in self.rotate_transform.ravel()
                                        ]
                                    )
                                },
                            }
                        )

                    d = dict(timepoint=t, setup=view_id, ViewTransform=view_transforms)

                    bdv_dict["ViewRegistrations"]["ViewRegistration"].append(d)

        return bdv_dict

    def stage_positions_to_affine_matrix(
        self, x: float, y: float, z: float, theta: float, f: Optional[float] = None
    ) -> npt.ArrayLike:
        """Convert stage positions to an affine matrix.

        Ignore theta, focus for now.

        Parameters
        ----------
        x : float
            The x position of the stage.
        y : float
            The y position of the stage.
        z : float
            The z position of the stage.
        theta : float
            The theta position of the stage.

        Returns
        -------
        npt.ArrayLike
            An affine matrix.
        """
        arr = np.eye(3, 4)

        # Set the transform positions
        xp, yp, zp = x / self.dx, y / self.dy, z / self.dz

        # Allow additional axes (e.g. f) to couple onto existing axes (e.g. z)
        # if they are both moving along the same physical dimension
        if self._coupled_axes is not None:
            for leader, follower in self._coupled_axes.items():
                if leader.lower() not in "xyz":
                    print(
                        f"Unrecognized coupled axis {leader}. "
                        "Not gonna do anything with this."
                    )
                    continue
                elif leader.lower() == "x":
                    xp += float(locals()[f"{follower.lower()}"]) / self.dx
                elif leader.lower() == "y":
                    yp += float(locals()[f"{follower.lower()}"]) / self.dy
                elif leader.lower() == "z":
                    zp += float(locals()[f"{follower.lower()}"]) / self.dz

        # Translation into pixels
        arr[:, 3] = [yp, xp, zp]

        # Rotation (theta pivots in the xz plane, about the y axis)
        # sin_theta, cos_theta = np.sin(theta), np.cos(theta)
        # arr[0,0], arr[2,2] = cos_theta, cos_theta
        # arr[0,2], arr[2,0] = sin_theta, -sin_theta

        return arr

    def affine_matrix_to_stage_positions(self, mat: npt.ArrayLike) -> tuple:
        """
        Convert affine matrix back into stage positions.

        Ignore theta, focus for now.

        Parameters
        ----------
        mat : npt.ArrayLike
            An affine matrix.

        Returns
        -------
        tuple
            A tuple of stage positions.
        """
        y, x, z = mat[:, 3] * np.array([self.dy, self.dx, self.dz])
        theta, f = None, None

        return x, y, z, theta, f

    def bdv_shear_transform(self):
        """Calculate the shear transform matrix.

        BDV-specific. Matrix provided is not (4,4), but (3,4).
        """
        if self.shear_data:
            self.shear_transform = affine_shear(
                dz=self.dz,
                dx=self.dx,
                dy=self.dy,
                dimension=self.shear_dimension,
                angle=self.shear_angle,
            )[:3]
        else:
            self.shear_transform = np.eye(3, 4)

    def bdv_rotate_transform(self):
        """Calculate the BDV rotation transform matrix.

        BDV-specific. Matrix provided is not (4,4), but (3,4).
        """
        if self.rotate_data:
            self.rotate_transform = affine_rotation(
                x=self.rotate_angle_x, y=self.rotate_angle_y, z=self.rotate_angle_z
            )[:3]
        else:
            self.rotate_transform = np.eye(3, 4)

    def parse_xml(self, root: Union[str, ET.Element]) -> tuple:
        """Parse a BigDataViewer XML file into our metadata format.

        Parameters
        ----------
        root : Union[str, ET.Element]
            The root of the XML tree.

        Returns
        -------
        tuple
            A tuple containing the file path, setups, and transforms.
        """

        # Open the file, if present
        if isinstance(root, str) and os.path.isfile(root):
            with open(root, "r") as fp:
                example = fp.read()
                root = ET.fromstring(example)

        if root.tag != "SpimData":
            raise NotImplementedError(f"Unknown format {root.tag} failed to load.")

        # Check if we are loading a BigDataViewer hdf5
        image_loader = root.find("SequenceDescription/ImageLoader")
        if image_loader.attrib["format"] not in ["bdv.hdf5", "bdv.n5"]:
            raise NotImplementedError(
                f"Unknown format {image_loader.attrib['format']} failed to load."
            )

        # Parse the file path
        base_path = root.find("BasePath")
        file = root.find("SequenceDescription/ImageLoader/hdf5")
        file_path = os.path.join(base_path.text, file.text)

        # Get setups. Each setup represents a visualisation data source in the viewer
        # that provides one image volume per timepoint
        setups = [
            x.text for x in root.findall("SequenceDescription/ViewSetups/ViewSetup/id")
        ]

        # Get channels, one per setup
        channels = list(
            set(
                [
                    x.text
                    for x in root.findall(
                        "SequenceDescription/ViewSetups/ViewSetup/attributes/channel"
                    )
                ]
            )
        )

        # Get number of positions, one per setup/channel
        self.positions = len(
            root.findall("SequenceDescription/ViewSetups/ViewSetup/attributes/tile")
        ) // len(channels)

        # Get image sizes in (x, y, z), one per setup
        sizes = [
            [int(y) for y in x.text.split()]
            for x in root.findall("SequenceDescription/ViewSetups/ViewSetup/size")
        ]

        # Get image voxel sizes (dx, dy, dz), one per setup
        voxel_sizes = [
            [float(y) for y in x.text.split()]
            for x in root.findall(
                "SequenceDescription/ViewSetups/ViewSetup/voxelSize/size"
            )
        ]

        # Get timepoints
        timepoint_type = root.find("SequenceDescription/Timepoints").attrib["type"]
        if timepoint_type != "range":
            raise NotImplementedError(
                f"Unknown format {timepoint_type} failed to load."
            )
        t_start = int(root.find("SequenceDescription/Timepoints/first").text)
        t_stop = int(root.find("SequenceDescription/Timepoints/last").text)
        timepoints = (t_start, t_stop + 1)

        # Get affine transformations, one per setup
        tt, ts = np.array(
            [
                [int(x.attrib["timepoint"]), int(x.attrib["setup"])]
                for x in root.findall("ViewRegistrations/ViewRegistration")
            ]
        ).T
        transforms = [
            np.array(x.text.split(), dtype=float).reshape(-1, 4)
            for x in root.findall(
                "ViewRegistrations/ViewRegistration/ViewTransform/affine"
            )
        ]

        # Set up metadata parameters
        self.dx, self.dy, self.dz = np.array(voxel_sizes).min(
            0
        )  # default to finest sampling
        self._multiposition = self.positions > 1
        self.shape_x, self.shape_y, self.shape_z = np.array(sizes).max(
            0
        )  # default to largest size captured
        self.shape_c = len(channels)
        self.shape_t = timepoints[1] - timepoints[0]

        return file_path, setups, transforms

    def write_xml(self, file_name: str, views: list) -> None:
        """Write BigDataViewer XML metadata.

        Parameters
        ----------
        file_name : str
            The file name of the file to be written.
        views : list
            A list of dictionaries containing metadata for each view.

        """

        return super().write_xml(
            file_name, file_type="bdv", root="SpimData", views=views
        )



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk
import logging
from pathlib import Path

# Third Party Imports

# Local Imports
from navigate.view.main_window_content.settings_notebook import SettingsNotebook
from navigate.view.main_window_content.display_notebook import CameraNotebook
from navigate.view.main_window_content.acquire_notebook import AcquireBar
from navigate.view.main_window_content.menus import Menubar
from navigate.view.custom_widgets.scrollbars import ScrolledFrame

# Logger Setup
p = __name__.split(".")[1]


class MainApp(ttk.Frame):
    """Creates the frame that will hold the GUI content, its parent is the main window
    or root Tk object

    A main window is created and passed to the mainapp class. This class will init as a
    frame then config the main window. It then creates a menubar using the menubar
    class.

    Adds the options for each file menu. It then sets up the frames, then grids the
    frames.

    Finally it uses the notebook classes to put them into the respective frames on the
    tk.Grid. Each of the notebook classes includes tab classes and inits those etc.

    The second parameter in each classes __init__ function is the parent.

    I used the name of the parent so that it would be easier to keep track of
    inheritances.

    Once you have the parent name you can look to the parents class in the class
    definition.

    For example for class Main_App(ttk.Frame) the parent to Main_App is a frame and its
    name is root. I also used the name of the class instead of self to make things
    easier to read. So for Main_App self is now mainapp.

    Placing the notebooks using tk.Grid. While the grid is called on each frame it
    is actually calling the main window since those are the parent to the frames.
    The labels have already been packed into each respective frame so can be ignored
    in the grid setup. This layout uses a 2x2 grid to start.

    1   2
    3   4
    5   6

    The above is the grid "spots" the left frame will take spots 3 & 5 while top
    right takes spot 4 and bottom right frame takes spot 6. Top frame will be
    spots 1 & 2
    """

    def __init__(self, root, *args, **kwargs):
        """Initiates the main application window

        Parameters
        ----------
        root : tk.Tk
            The main window of the application
        *args
            Variable length argument list
        **kwargs
            Arbitrary keyword arguments
        """

        # Inits this class as a frame subclass with the root as its parent
        #: ScrolledFrame: The scrollable version of the main frame for the application
        self.scroll_frame = ScrolledFrame(root)
        self.scroll_frame.grid(row=0, column=0, sticky=tk.NSEW)

        ttk.Frame.__init__(self, self.scroll_frame.interior, *args, **kwargs)

        # Initialize Logger
        #: logging.Logger: The logger for this class
        self.logger = logging.getLogger(p)

        # This starts the main window config, and makes sure that any child
        # widgets can be resized with the window
        #: tk.Tk: The main window of the application
        self.root = root
        self.root.title("navigate")

        # keep icons relative to view directory structure
        view_directory = Path(__file__).resolve().parent
        try:
            photo_image = view_directory.joinpath("icon", "mic.png")
            self.root.iconphoto(True, tk.PhotoImage(file=photo_image))
        except tk.TclError:
            pass
        self.root.resizable(True, True)
        self.root.geometry("")
        tk.Grid.columnconfigure(root, "all", weight=1)
        tk.Grid.rowconfigure(root, "all", weight=1)

        # Creating and linking menu to main window/app
        #: Menubar: The menu bar for the application
        self.menubar = Menubar(root)

        # Top Frame Acquire Bar
        #: ttk.Frame: The top frame of the application
        self.top_frame = ttk.Frame(self)

        # Left Frame Notebook 1 setup
        #: ttk.Frame: The left frame of the application
        self.frame_left = ttk.Frame(self)

        # Top right Frame Notebook 2 setup
        #: ttk.Frame: The top right frame of the application
        self.frame_top_right = ttk.Frame(self)

        # Bottom right Frame Notebook 3 setup
        #: ttk.Frame: The bottom right frame of the application
        self.frame_bottom_right = ttk.Frame(self)

        # Grid out foundational frames
        self.grid(column=0, row=0, sticky=tk.NSEW)
        self.top_frame.grid(
            row=0, column=0, columnspan=2, sticky=tk.NSEW, padx=3, pady=3
        )
        self.frame_left.grid(row=1, column=0, rowspan=2, sticky=tk.NSEW, padx=3, pady=3)
        self.frame_top_right.grid(row=1, column=1, sticky=tk.NSEW, padx=3, pady=3)
        self.frame_bottom_right.grid(row=2, column=1, sticky=tk.NSEW, padx=3, pady=3)

        # Putting Notebooks into frames, tabs are held within the class of each
        # notebook
        #: SettingsNotebook: The settings notebook for the application
        self.settings = SettingsNotebook(self.frame_left, self.root)
        #: CameraNotebook: The camera notebook for the application
        self.camera_waveform = CameraNotebook(self.frame_top_right, self.root)
        #: AcquireBar: The acquire bar for the application
        self.acqbar = AcquireBar(self.top_frame, self.root)



""" Hardware devices. """

from .daq.daq_synthetic import SyntheticDAQ  # noqa
from .camera.camera_synthetic import SyntheticCamera  # noqa
from .filter_wheel.filter_wheel_synthetic import SyntheticFilterWheel  # noqa
from .galvo.galvo_synthetic import SyntheticGalvo  # noqa
from .remote_focus.remote_focus_synthetic import SyntheticRemoteFocus  # noqa
from .shutter.laser_shutter_synthetic import SyntheticShutter  # noqa
from .stages.stage_synthetic import SyntheticStage  # noqa
from .zoom.zoom_synthetic import SyntheticZoom  # noqa
from .lasers.laser_synthetic import SyntheticLaser  # noqa
from .mirrors.mirror_synthetic import SyntheticMirror  # noqa



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
import logging
from tkinter import ttk
import tkinter as tk
from tkinter.scrolledtext import ScrolledText

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedCombobox, ValidatedEntry
from navigate.model.data_sources import FILE_TYPES

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class AcquirePopUp:
    """Class creates the popup that is generated when the Acquire button is pressed and
    Save File checkbox is selected."""

    def __init__(self, root, *args, **kwargs):
        """Initialize the AcquirePopUp class

        Parameters
        ----------
        root : tk.Tk
            The root window
        *args
            Variable length argument list
        **kwargs
            Arbitrary keyword arguments
        """
        # Creating popup window with this name and size/placement, PopUp is a Toplevel
        # window
        #: PopUp: The popup window
        self.popup = PopUp(
            root, "File Saving Dialog", "430x530+320+180", transient=True
        )

        # Storing the content frame of the popup, this will be the parent of the widgets
        content_frame = self.popup.get_frame()

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        # Dictionary for all the variables
        #: dict: Dictionary of all the variables
        self.inputs = {}
        #: dict: Dictionary of all the buttons
        self.buttons = {}

        # Label for entries
        #: ttk.Label: Label for the entries
        self.entries_label = ttk.Label(
            content_frame, text="Please fill out the fields below"
        )
        self.entries_label.grid(row=0, column=0, columnspan=2, sticky=tk.NSEW, pady=5)

        # Creating Entry Widgets
        entry_names = [
            "root_directory",
            "user",
            "tissue",
            "celltype",
            "label",
            "solvent",
            "file_type",
            "misc",
        ]

        entry_labels = [
            "Root Directory",
            "User",
            "Tissue Type",
            "Cell Type",
            "Label",
            "Solvent",
            "File Type",
            "Notes",
        ]

        # Loop for each entry and label
        for i in range(len(entry_names)):
            if entry_names[i] == "misc":
                self.inputs[entry_names[i]] = LabelInput(
                    parent=content_frame,
                    label=entry_labels[i],
                    input_class=ScrolledText,
                    input_args={"wrap": tk.WORD, "width": 40, "height": 10},
                )
            elif entry_names[i] == "file_type":
                self.inputs[entry_names[i]] = LabelInput(
                    parent=content_frame,
                    label=entry_labels[i],
                    input_class=ValidatedCombobox,
                    input_var=tk.StringVar(),
                )
                self.inputs[entry_names[i]].set_values(tuple(FILE_TYPES))
                self.inputs[entry_names[i]].set("TIFF")

            elif entry_names[i] == "solvent":
                self.inputs[entry_names[i]] = LabelInput(
                    parent=content_frame,
                    label=entry_labels[i],
                    input_class=ValidatedCombobox,
                    input_var=tk.StringVar(),
                )
                self.inputs[entry_names[i]].set_values(
                    ("BABB", "Water", "CUBIC", "CLARITY", "uDISCO", "eFLASH")
                )
                self.inputs[entry_names[i]].set("BABB")

            else:
                self.inputs[entry_names[i]] = LabelInput(
                    parent=content_frame,
                    label=entry_labels[i],
                    input_class=ttk.Entry,
                    input_var=tk.StringVar(),
                    input_args={"width": 50},
                )
            self.inputs[entry_names[i]].grid(
                row=i + 1, column=0, columnspan=2, sticky=tk.NSEW, padx=5
            )
            self.inputs[entry_names[i]].label.grid(padx=(0, 20))

        # Formatting
        self.inputs["user"].widget.grid(padx=(53, 0))
        self.inputs["tissue"].widget.grid(padx=(16, 0))
        self.inputs["celltype"].widget.grid(padx=(28, 0))
        self.inputs["label"].widget.grid(padx=(48, 0))
        self.inputs["misc"].widget.grid(padx=(24, 0))

        # Done and Cancel Buttons
        self.buttons["Cancel"] = ttk.Button(content_frame, text="Cancel Acquisition")
        self.buttons["Cancel"].grid(row=9, column=0, padx=5, sticky=tk.NSEW)

        self.buttons["Done"] = ttk.Button(content_frame, text="Acquire Data")
        self.buttons["Done"].grid(row=9, column=1, padx=5, sticky=tk.NSEW)

    def get_variables(self):
        """Get the variables of the popup

        This function returns a dictionary of all the variables that are tied to each
        widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            Dictionary of all the variables
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """Get the widgets of the popup

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            Dictionary of all the widgets
        """
        return self.inputs

    def get_buttons(self):
        """Get the buttons of the popup

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        dict
            Dictionary of all the buttons
        """
        return self.buttons



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Laser Base Class
"""
import logging

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class LaserBase:
    """Laser Base Class"""

    def __init__(self, microscope_name, device_connection, configuration, laser_id):
        """Initialize Laser Base Class

        Parameters
        ----------
        microscope_name : str
            Name of the microscope
        device_connection : str
            Connection string for the device
        configuration : dict
            Configuration dictionary
        laser_id : int
            Laser ID
        """

        #: dict: Configuration dictionary
        self.configuration = configuration

        #: str: Name of the microscope
        self.microscope_name = microscope_name

        #: int: Laser ID
        self.device_config = configuration["configuration"]["microscopes"][
            microscope_name
        ]["lasers"][laser_id]

    def set_power(self, laser_intensity):
        """Set laser power

        Parameters
        ----------
        laser_intensity : int
            Laser intensity
        """
        pass

    def turn_on(self):
        """Turn on the laser"""
        pass

    def turn_off(self):
        """Turn off the laser"""
        pass

    def close(self):
        """
        Close the laser before exit.
        """
        pass

    def initialize_laser(self):
        """Initialize lasers."""
        pass



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging

# Third Party Imports
import numpy.typing as npt

# Local Imports
from multiprocessing.managers import DictProxy


class DataSource:
    def __init__(self, file_name: str = "", mode: str = "w") -> None:
        """Base class for data sources, which can be of arbitrary file type.

        This implements read and write methods for accessing each data source.
        We expect to open the file for read or write during initialization.

        file_name  stores the name of the file to read/write from
        data       stores a pointer to the actual data in the data source and
        metdata    stores a pointer to the metadata.

        Concept and some of the code borrowed from python-microscopy
        (https://github.com/python-microscopy/python-microscopy/).

        Parameters
        ----------
        file_name : str
            Name of the file to read/write from.
        mode : str
            Mode to open the file in. Can be 'r' or 'w'.
        """
        #: logging.Logger: Logger for this class.
        self.logger = logging.getLogger(__name__.split(".")[1])
        #: str: Name of the file to read/write from.
        self.file_name = file_name
        if not hasattr(self, "metadata"):
            #: npt.ArrayLike: Pointer to the metadata.
            self.metadata = None  # Expect a metadata object
        #: str: Mode to open the file in. Can be 'r' or 'w'.
        self._mode = None
        #: bool: Has the data source been closed?
        self._closed = True
        #: int: Number of bits per pixel.
        self.bits = 16

        #: float: Pixel size in x dimension (microns).
        #: float: Pixel size in y dimension (microns).
        #: float: Pixel size in z dimension (microns).
        self.dx, self.dy, self.dz = 1, 1, 1  # pixel sizes (um)

        #: float: The time interval between frames (seconds).
        self.dt = 1  # time displacement (s)

        #: float: Step size between channels (always 1)
        self.dc = 1  # step size between channels, should always be 1

        #: int: Size of the data source in x dimension.
        #: int: Size of the data source in y dimension.
        #: int: Size of the data source in z dimension.
        #: int: Size of the data source in t dimension.
        #: int: Size of the data source in c dimension.
        self.shape_x, self.shape_y, self.shape_z, self.shape_t, self.shape_c = (
            1,
            1,
            1,
            1,
            1,
        )
        #: int: Number of positions in the data source.
        self.positions = 1

        #: str: Mode to open the file in. Can be 'r' or 'w'.
        self.mode = mode

        #: int: Current frame number.
        self._current_frame = 0

    @property
    def nbytes(self) -> int:
        """Getter for the size of this data source in bytes."

        Does not account for pyramidal data sources. That process is handled by the
        bdv_data_source class, which is a child of this class.

        Returns
        -------
        int
            Size of this data source in bytes.
        """
        total_bits = (
            self.shape_x
            * self.shape_y
            * self.shape_z
            * self.shape_t
            * self.shape_c
            * self.positions
            * self.bits
        )
        total_bytes = total_bits // 8
        return total_bytes

    @property
    def mode(self) -> str:
        """Getter for the mode of the data source.

        Returns
        -------
        str
            Mode of the data source.
        """
        return self._mode

    @mode.setter
    def mode(self, mode_str) -> None:
        """Setter for the mode of the data source.

        Parameters
        ----------
        mode_str : str
            Mode to set the data source to. Can be 'r' or 'w'.
        """
        if mode_str == self._mode:
            return
        if mode_str in ["r", "w"]:
            self._mode = mode_str
        else:
            self.logger.warning(f"Unknown mode {mode_str}. Setting to 'r'.")
            self._mode = "r"
        self._mode_checks()

    @property
    def data(self) -> npt.ArrayLike:
        """Return array representation of data stored in data source.

        Returns
        -------
        npt.ArrayLike
            Array representation of data stored in data source.

        Raises
        ------
        NotImplementedError
            If not implemented in a derived class.
        """
        raise NotImplementedError("Implemented in a derived class.")

    @property
    def voxel_size(self) -> tuple:
        """Getter for the voxel size

        Returns
        -------
        tuple
            Voxel size in x, y, z dimensions.
        """
        return (self.dx, self.dy, self.dz)

    @property
    def shape(self) -> tuple:
        """Getter for the shape as XYCZT.

        Returns
        -------
        tuple
            Shape of the data source in XYCZT format.
        """
        return (self.shape_x, self.shape_y, self.shape_c, self.shape_z, self.shape_t)

    def set_metadata_from_configuration_experiment(
        self, configuration: DictProxy
    ) -> None:
        """Sets the metadata from according to the microscope configuration.

        Parameters
        ----------
        configuration : DictProxy
            Configuration experiment.
        """

        self.metadata.configuration = configuration
        self.get_shape_from_metadata()

    def set_metadata(self, metadata_config: dict) -> None:
        """Sets the metadata
        
        Parameters
        ----------
        metadata_config : dict
            shape configuration: "c", "z", "t", "p", "is_dynamic", "per_stack"
        """
        self.metadata.set_from_dict(metadata_config)
        self.get_shape_from_metadata()

    def get_shape_from_metadata(self):
        """Get the shape of the data source from the metadata."""
        # pull new values from the metadata
        self.dx, self.dy, self.dz = self.metadata.voxel_size
        (
            self.shape_x,
            self.shape_y,
            self.shape_c,
            self.shape_z,
            self.shape_t,
        ) = self.metadata.shape
        self.positions = self.metadata.positions

    def _cztp_indices(self, frame_id: int, per_stack: bool = True) -> tuple:
        """Figure out where we are in the stack from the frame number.

        per_stack indicates if we move through z (per_stack=True) or c fastest.
        we move through positions slower than z or c. we move through time slower
        than z, c or p.

        Parameters
        ----------
        frame_id : int
            Frame number in the stack.
        per_stack : bool
            Are we acquiring images along z before c? See experiment.yml.

        Returns
        -------
        c : int
            Index of channel
        z : int
            Index of z position
        t : int
            Index of time position
        p : int
            Index of multi-position position.
        """
        # If z-stacking, if multi-position
        if self.shape_z > 1:
            # We're z-stacking, make z, c vary faster than t
            if per_stack:
                c = (frame_id // self.shape_z) % self.shape_c
                z = frame_id % self.shape_z
            else:
                c = frame_id % self.shape_c
                z = (frame_id // self.shape_c) % self.shape_z

            t = (frame_id // (self.shape_c * self.shape_z)) % self.shape_t
            p = frame_id // (self.shape_c * self.shape_z * self.shape_t)

            # NOTE: Uncomment this if we want positions to vary faster than time
            # t = frame_id // (self.shape_c * self.shape_z * self.positions)
            # p = (frame_id // (self.shape_c * self.shape_z)) % self.positions

        else:
            # Timepoint acquisition, only c varies faster than t
            c = frame_id % self.shape_c
            t = (frame_id // self.shape_c) % self.shape_t
            z = (frame_id // (self.shape_c * self.shape_t)) % self.shape_z
            p = frame_id // (self.shape_c * self.shape_t * self.shape_z)

        return c, z, t, p

    def _check_shape(self, max_frame: int = 0, per_stack: bool = True):
        """Check if we've closed this prior to completion.

        Parameters
        ----------
        max_frame : int
            Maximum frame number.
        per_stack : bool
            Acquisition mode. Either per_stack of per_channel.
        """

        # Check if we've closed this prior to completion
        c, z, t, p = self._cztp_indices(max_frame, per_stack)
        # print(f"max_frame: {max_frame} c: {c} z: {z} t: {t} p: {p}")
        # print(f"XYCZTP: {self.shape} {self.positions}")
        if (
            (z < (self.shape_z - 1))
            or (c < (self.shape_c - 1))
            or (t < (self.shape_t - 1))
            or (p < (self.positions - 1))
        ):
            # If we have, update our shape accordingly
            maxc, maxz, maxt, maxp = 0, 0, 0, 0
            for idx in range(max_frame + 1):
                c, z, t, p = self._cztp_indices(idx, per_stack)
                maxc = max(maxc, c)
                maxz = max(maxz, z)
                maxt = max(maxt, t)
                maxp = max(maxp, p)
            self.shape_c, self.shape_z = maxc + 1, maxz + 1
            self.shape_t, self.positions = maxt + 1, maxp + 1
            if self.metadata is not None:
                self.metadata.shape_c, self.metadata.shape_z = maxc + 1, maxz + 1
                self.metadata.shape_t, self.metadata.positions = maxt + 1, maxp + 1
        # print(
        #     f"result c: {self.shape_c} z: {self.shape_z} "
        #     f"t: {self.shape_t} p: {self.positions}"
        # )

    def _mode_checks(self) -> None:
        """Run additional checks after setting the mode."""
        pass

    def write(self, data: npt.ArrayLike, **kw) -> None:
        """Write data to file.

        Parameters
        ----------
        data : npt.ArrayLike
            Data to write to file.
        **kw : dict
            Additional keyword arguments.

        Raises
        ------
        NotImplementedError
            If not implemented in a derived class.
        """
        raise NotImplementedError("Implemented in a derived class.")

    def read(self) -> None:
        """Read data from file.

        Raises
        ------
        NotImplementedError
            If not implemented in a derived class.
        """
        raise NotImplementedError("Implemented in a derived class.")

    def close(self) -> None:
        """Clean up any leftover file pointers, etc."""
        pass

    def __del__(self):
        """Destructor"""
        if not self._closed:
            self.close()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports

# Third party imports

# Local application imports
from navigate.model.features.autofocus import Autofocus


class CalculateFocusRange:
    """CalculateFocusRange class for determining the focus range using autofocus.

    This class provides functionality to calculate the focus range of a microscope
    using autofocus measurements.

    Notes:
    ------
    - This class is used to calculate the focus range of a microscope by utilizing
    autofocus measurements. It provides methods for different stages of the
    calculation process.

    - The `config_table` attribute is a dictionary that defines the configuration for
    each stage of the calculation process, including initialization, main execution,
    and finalization steps.

    - The autofocus measurements are performed using an instance of the `Autofocus`
    class, which is initialized with the same `model` object used for focus range
    calculation.

    - The focus range calculation involves signal acquisition and data acquisition
    stages, which are controlled using the provided methods.

    - The `CalculateFocusRange` class can be used to automate the process of
    determining the optimal focus range for microscopy experiments.
    """

    def __init__(self, model):
        """Initialize the CalculateFocusRange class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for focus range calculation.
        """
        #: MicroscopeModel: The microscope model object used for focus range
        self.model = model
        #: Autofocus: The autofocus object used for focus range calculation.
        self.autofocus = Autofocus(model)
        #: dict: A dictionary that defines the configuration for each stage of the
        # calculation process, including initialization, main execution, and
        # finalization steps.
        self.config_table = {
            "signal": {
                "init": self.pre_func_signal,
                "main": self.in_func_signal,
                "end": self.end_func_signal,
            },
            "data": {
                "init": self.pre_func_data,
                "main": self.in_func_data,
                "end": self.end_func_data,
            },
            "node": {"node_type": "multi-step", "device_related": True},
        }

    def pre_func_signal(self):
        """Prepare for the signal acquisition stage.

        This method sets up initial values and configurations before the signal
        acquisition stage.
        """
        self.model.active_microscope.current_channel = 0
        self.model.active_microscope.prepare_next_channel()
        self.autofocus.pre_func_signal()
        #: int: The number of autofocus measurements performed.
        self.autofocus_count = 0
        #: float: The starting z position of the focus range.
        self.focus_start_pos = None
        #: float: The ending z position of the focus range.
        self.focus_end_pos = None

        # get current z pos, calculate last z pos in a stack
        stage_pos = self.model.get_stage_position()
        #: float: The current z position of the stage.
        self.current_z_pos = stage_pos["z_pos"]
        #: float: The last z position of the stage in a stack.
        self.last_z_pos = self.current_z_pos + float(
            self.model.configuration["experiment"]["MicroscopeState"]["end_position"]
        )

    def in_func_signal(self):
        """Perform actions during the signal acquisition stage.

        This method is responsible for carrying out actions during the signal
        acquisition stage, such as performing autofocus measurements and calculating
        focus-related parameters.
        """
        if self.autofocus_count == 0:
            self.focus_start_pos = self.autofocus.in_func_signal()
        else:
            self.focus_end_pos = self.autofocus.in_func_signal()

            # calculate the slope and save it
            if self.focus_end_pos:
                microscope_state = self.model.configuration["experiment"][
                    "MicroscopeState"
                ]
                microscope_state["end_focus"] = (
                    float(microscope_state["start_focus"])
                    + self.focus_end_pos
                    - self.focus_start_pos
                )

    def end_func_signal(self):
        """Finalize the signal acquisition stage.

        This method finalizes the signal acquisition stage and checks if additional
        steps are needed based on the results of autofocus measurements.

        Returns:
        -------
        bool
            A boolean value indicating whether additional steps are required.
        """
        r = self.autofocus.end_func_signal()
        if r:
            self.autofocus_count += 1
            if self.autofocus_count == 1:
                # move one z step
                # TODO: should the focus move at the same time?
                self.model.move_stage({"z_abs": self.last_z_pos}, wait_until_done=True)
                self.autofocus.pre_func_signal()
        return self.autofocus_count >= 2

    def pre_func_data(self):
        """Prepare for the data acquisition stage, if applicable.

        This method prepares for the data acquisition stage, which may involve
        setting up configurations or performing actions before data acquisition begins.
        """
        self.autofocus.pre_func_data()

    def in_func_data(self, frame_ids=[]):
        """Perform actions during the data acquisition stage, possibly for specific
        frame IDs.

        This method performs actions during the data acquisition stage, which may
        include data
        acquisition for specific frame IDs if provided.

        Parameters:
        ----------
        frame_ids : list, optional
            A list of frame IDs for which data acquisition should be performed.
            Default is an empty list.
        """
        self.autofocus.in_func_data(frame_ids)

    def end_func_data(self):
        """Finalize the data acquisition stage, if applicable.

        This method finalizes the data acquisition stage, and it may involve
        additional actions or
        checks.

        Returns:
        -------
        bool
            A boolean value indicating whether additional steps are required.
        """
        r = self.autofocus.end_func_data()
        if r:
            self.autofocus.pre_func_data()

        return r and self.autofocus_count >= 2



""" File type specific data sources. """

FILE_TYPES = ["TIFF", "OME-TIFF", "H5", "N5"]


def get_data_source(file_type):
    """Get the data source class for the given file type.

    Parameters
    ----------
    file_type : str
        File type to get the data source for.

    Returns
    -------
    DataSource
        Data source class for the given file type.

    Examples
    --------
    >>> get_data_source('tif')
    <class 'navigate.model.data_sources.tif.TifDataSource'>
    """

    if (file_type == "TIFF") or (file_type == "OME-TIFF"):
        from .tiff_data_source import TiffDataSource

        return TiffDataSource

    elif (file_type == "H5") or file_type == ("N5"):
        from .bdv_data_source import BigDataViewerDataSource

        return BigDataViewerDataSource

    else:
        raise NotImplementedError(f"Unknown file type {file_type}. Cannot open.")



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging

# Third Party Imports
# you shouldn't need to call this as self.controller is the camera object
# and should have everything needed
from ctypes import *  # noqa
import numpy as np

# Local Imports
from navigate.model.devices.camera.camera_base import CameraBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class PhotometricsKinetix(CameraBase):
    """Photometrics Kinetix camera class.

    This class is the interface between the rest of the microscope code and the
    Photometrics API.
    """

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize the Photometrics class.

        Parameters
        ---------
        microscope_name : str
                Name of microscope in configuration
        device_connection : object
                Hardware device to connect to. will be saved in camera_controller
        configuration : multiprocesing.managers.DictProxy
                Global configuration of the microscope
        -------

        """
        super().__init__(microscope_name, device_connection, configuration)

        #: int: Exposure Time in milliseconds
        self._exposuretime = 20
        #: int: Scan Mode (0 = Normal, 1 = ASLM)
        self._scanmode = 0
        #: int: Scan Delay
        self._scandelay = 1
        #: int: Number of frames
        self._numberofframes = 100
        #: obj: Data Buffer
        self._databuffer = None
        #: int: Number of frames received
        self._frames_received = 0
        #: float: Unit for line delay
        self._unitforlinedelay = 0.00375  # 3.75  us for dynamic mode kinetix
        #: list: Frame IDs
        self._frame_ids = []
        #: int: Max image width
        self.max_image_width = self.camera_controller.sensor_size[0]
        #: int: Max image height
        self.max_image_height = self.camera_controller.sensor_size[1]
        #: dict: Camera parameters
        self.camera_parameters["x_pixels"] = self.max_image_width
        self.camera_parameters["y_pixels"] = self.max_image_height

        # todo: complete first init of parameters to default values

        # Values are pulled from the CameraParameters section of the
        # configuration.yml file.
        # Exposure time converted here from milliseconds to seconds.
        self.set_sensor_mode("Normal")
        self.camera_controller.binning = 1
        # not implemented: readout_speed, defect_correct_mode
        self.camera_controller.exp_mode = "Edge Trigger"
        self.camera_controller.prog_scan_dir = 0
        # self.camera_controller.speed_table_index = 1 # 1 for 100 MHz
        self.camera_controller.readout_port = 0
        self.camera_controller.gain = 1
        #
        # self.camera_controller.set_property_value("trigger_active",
        #                                           self.camera_parameters[
        #                                           'trigger_active'])
        # self.camera_controller.set_property_value("trigger_mode",
        #                                           self.camera_parameters[
        #                                           'trigger_mode'])
        # self.camera_controller.set_property_value("trigger_polarity",
        #                                           self.camera_parameters[
        #                                           \'trigger_polarity'])
        # self.camera_controller.set_property_value("trigger_source",
        #                                           self.camera_parameters[
        #                                           'trigger_source'])
        # # DCAM_IDPROP_IMAGE_WIDTH/HEIGHT is readonly
        # self.camera_controller.set_property_value("image_height",
        #                                            self.camera_parameters[
        #                                            'y_pixels'])
        # self.camera_controller.set_property_value("image_width",
        #                                            self.camera_parameters[
        #                                            'x_pixels'])

        logger.info("Photometrics Initialized")

    def __del__(self):
        """Delete PhotometricsKinetix object."""
        if hasattr(self, "camera_controller"):
            self.camera_controller.close()
            # pvc.uninit_pvcam()
            print("camera closed")
        logger.info("PhotometricsKinetix Shutdown")

    @property
    def serial_number(self):
        """Get Camera Serial Number

        Returns
        -------
        serial_number : str
            Serial number for the camera.
        """
        # return self.camera_controller._serial_number
        ser_no = self.camera_controller.serial_no
        return ser_no

    def report_settings(self):
        """Print Camera Settings."""
        # todo: complete param recording
        print("sensor_mode: " + str(self.camera_controller.prog_scan_mode))
        print("binning: " + str(self.camera_controller.binning))
        print("readout_speed" + str(self.camera_controller.readout_time))
        print("trigger_active")
        print("trigger_mode")
        print("trigger_polarity")
        print("trigger_source")
        print("internal_line_interval")
        print("sensor size" + str(self.camera_controller.sensor_size))
        print("image_height and width" + str(self.x_pixels) + ", " + str(self.y_pixels))

        print("exposure_time" + str(self._exposuretime))

    def close_camera(self):
        """Close Photometrics Kinetix Camera"""
        self.camera_controller.close()

    def set_sensor_mode(self, mode):
        """Set Photometrics sensor mode

        Can be normal or programmable scan mode (e.g., ASLM).

        Parameters
        ----------
        mode : str
            'Normal (static)' or 'Light-Sheet (ASLM)'
        """
        modes_dict = {"Normal": 0, "Light-Sheet": 1}
        if mode in modes_dict:
            self.camera_controller.prog_scan_mode = modes_dict[mode]
            self._scanmode = modes_dict[mode]
        else:
            print("Camera mode not supported" + str(modes_dict[mode]))
            logger.info("Camera mode not supported" + str(modes_dict[mode]))

    def set_readout_direction(self, mode):
        """Set Photometrics readout direction.

        Parameters
        ----------
        mode : str
            'Top-to-Bottom', 'Bottom-to-Top', 'Alternate'
            Scan direction options: {'Down': 0, 'Up': 1, 'Down/Up Alternate': 2}

        """
        print("available scan directions on camera are:")
        print(str(self.camera_controller.prog_scan_dirs))

        if mode == "Top-to-Bottom":
            #  'Down' readout direction
            self.camera_controller.prog_scan_dir = 0
        elif mode == "Bottom-to-Top":
            #  'Up' readout direction
            self.camera_controller.prog_scan_dir = 1
        elif mode == "Alternate":
            self.camera_controller.prog_scan_dir = 2
        else:
            print("Camera readout direction not supported")
            logger.info("Camera readout direction not supported")

    def calculate_readout_time(self):
        """Calculate duration of time needed to readout an image.

        Calculates the readout time and maximum frame rate according to the camera
        configuration settings.

        Warn
        ----
            Not implemented. Currently hard-coded for Hamamatsu Orca Flash 4.0.

        Returns
        -------
        readout_time : float
            Duration of time needed to readout an image.
        max_frame_rate : float
            Maximum framerate for a given camera acquisition mode.
        """

        # todo
        h = 3.75 * 10**-6  # Readout timing constant
        # h = self.camera_controller.get_property_value("readout_time")
        vn = self.y_pixels
        exposure_time = self._exposuretime
        # trigger_source = self.camera_controller.get_property_value('trigger_source')
        # trigger_active = self.camera_controller.get_property_value('trigger_active')
        #
        if self._scanmode == 0:  # normal/static light-sheet
            max_frame_rate = 1 / ((vn + 10) * h + exposure_time)
            readout_time = exposure_time - ((vn + 10) * h + exposure_time)
        else:
            # todo: not sure if these equations are correct
            max_frame_rate = 1 / ((vn + 10) * h * self._scandelay + exposure_time)
            readout_time = exposure_time - (
                (vn + 10) * h * self._scandelay + exposure_time
            )

            #
        #     #  Area sensor mode operation
        #     if trigger_source == 1:
        #         # Internal Trigger Source
        #         max_frame_rate = 1 / ((vn / 2) * h)
        #         readout_time = exposure_time - ((vn / 2) * h)
        #
        #     if trigger_active == 1 or 2:
        #         #  External Trigger Source
        #         #  Edge == 1, Level == 2
        #         max_frame_rate = 1 / ((vn / 2) * h + exposure_time + 10 * h)
        #         readout_time = exposure_time - ((vn / 2) * h + exposure_time + 10 * h)
        #
        #     if trigger_active == 3:
        #         #  External Trigger Source
        #         #  Synchronous Readout == 3
        #         max_frame_rate = 1 / ((vn / 2) * h + 5 * h)
        #         readout_time = exposure_time - ((vn / 2) * h + 5 * h)
        #
        # if sensor_mode == 12:
        #     #  Progressive sensor mode operation
        #     max_frame_rate = 1 / (exposure_time + (vn + 10) * h)
        #     readout_time = exposure_time - 1 / (exposure_time + (vn + 10) * h)
        #
        return readout_time, max_frame_rate

    def set_exposure_time(self, exposure_time):
        """Set Photometrics exposure time.

        Units of the Photometrics API are in seconds.
        All of our units are in milliseconds.

        Parameters
        ----------
        exposure_time : float
            Exposure time in milliseconds.

        Returns
        -------
        exposure_time : float
            Exposure time in seconds.
        """
        self._exposuretime = exposure_time
        return exposure_time

    def set_line_interval(self, line_interval_time):
        """Set Photometrics line interval.

        Parameters
        ----------
        line_interval_time : float
            Line interval duration.
        """
        # todo calculate line delay from scandelay
        self._scandelay = line_interval_time

    def calculate_light_sheet_exposure_time(
        self, full_chip_exposure_time, shutter_width
    ):
        """Convert normal mode exposure time to light-sheet mode exposure time.
        Calculate the parameters for an ASLM acquisition

        Parameters
        ----------
        full_chip_exposure_time : float
            Normal mode exposure time.
        shutter_width : int

        Returns
        -------
        exposure_time : float
            Light-sheet mode exposure time.
        camera_line_interval : float
            HamamatsuOrca line interval duration.
        """
        linedelay = self._unitforlinedelay  # 10.16us
        nbrows = self.y_pixels
        ASLM_scanWidth = 70

        ASLM_lineExposure = int(
            np.ceil(full_chip_exposure_time / (1 + nbrows / ASLM_scanWidth))
        )
        ASLM_line_delay = (
            int(
                np.ceil(
                    (full_chip_exposure_time - ASLM_lineExposure) / (nbrows * linedelay)
                )
            )
            - 1
        )
        ASLM_acquisition_time = (
            (ASLM_line_delay + 1) * nbrows * linedelay
            + ASLM_lineExposure
            + (ASLM_line_delay + 1) * linedelay
        )

        self.camera_parameters["line_interval"] = ASLM_lineExposure

        self._exposuretime = ASLM_lineExposure
        self._scandelay = ASLM_line_delay
        print(
            "ASLM parameters are: {} exposure time, and {} line delay factor, {} "
            "total acquisition time for {} scan width".format(
                ASLM_lineExposure,
                ASLM_line_delay,
                ASLM_acquisition_time,
                ASLM_scanWidth,
            )
        )

        return ASLM_lineExposure, ASLM_line_delay

    def _calculate_ASLMparameters(self, desired_exposuretime):
        """Calculate the parameters for an ASLM acquisition

        Parameters
        ----------
        desired_exposuretime : float
            Exposure time in milliseconds.

        Returns
        -------
        exposure_time : float
            Light-sheet mode exposure time.

        Warn
        ----
            Not implemented. No code in this function.
        """

    def set_binning(self, binning_string):
        """Set HamamatsuOrca binning mode.

        Parameters
        ----------
        binning_string : str
            Desired binning properties (e.g., '1x1', '2x2',
            '4x4', '8x8', '16x16', '1x2', '2x4')

        Returns
        -------
        result: bool
            True if binning was set successfully, False otherwise.

        """
        binning_dict = {
            "1x1": 1,
            "2x2": 2,
            "4x4": 4,
        }
        if binning_string not in binning_dict.keys():
            logger.debug(f"can't set binning to {binning_string}")
            print(f"can't set binning to {binning_string}")
            return False
        self.camera_controller.binning = binning_dict[binning_string]
        idx = binning_string.index("x")
        #: int: Binning in x direction
        self.x_binning = int(binning_string[:idx])
        #: int: Binning in y direction
        self.y_binning = int(binning_string[idx + 1 :])
        self.x_pixels = int(self.x_pixels / self.x_binning)
        self.y_pixels = int(self.y_pixels / self.y_binning)
        # should update experiment in controller side
        # self.configuration['experiment']['CameraParameters']['camera_binning'] = str(
        # self.x_binning) + 'x' + str(self.y_binning)
        return True

    def set_ROI(self, roi_height=3200, roi_width=3200):
        """Change the size of the active region on the camera.

        Parameters
        ----------
        roi_height : int
            Height of active camera region.
        roi_width : int
            Width of active camera region.
        """
        # Get the Maximum Number of Pixels from the Configuration File
        camera_height = self.camera_parameters["y_pixels"]
        camera_width = self.camera_parameters["x_pixels"]

        if (
            roi_height > camera_height
            or roi_width > camera_width
            or roi_height % 2 == 1
            or roi_width % 2 == 1
        ):
            logger.debug(f"can't set roi to {roi_width} and {roi_height}")
            return False

        # Calculate Location of Image Edges
        roi_top = (camera_height - roi_height) / 2
        roi_bottom = roi_top + roi_height - 1
        roi_left = (camera_width - roi_width) / 2

        if roi_top % 2 != 0 or roi_bottom % 2 == 0:
            logger.debug(f"can't set ROI to {roi_width} and {roi_height}")
            return False

        # Set ROI
        self.camera_controller.set_roi(roi_left, roi_top, roi_width, roi_height)
        # self.x_pixels, self.y_pixels = self.camera_controller.shape()

        logger.info(f"Photometrics ROI shape, {self.camera_controller.shape()}")

        return self.x_pixels == roi_width and self.y_pixels == roi_height

    def initialize_image_series(self, data_buffer=None, number_of_frames=100):
        """Initialize Photometrics image series. This is for starting stacks etc.

        Parameters
        ----------
        data_buffer : int
            Size of the data to buffer.  Default is None.
        number_of_frames : int
            Number of frames.  Default is 100.
        """
        print(self._exposuretime)
        self.camera_controller.readout_port = 0
        self.camera_controller.speed_table_index = 0
        self.camera_controller.gain = 1
        self._exposuretime = 20

        # set following parameters if in programmable scan mode (ASLM)
        self._scanmode = self.camera_controller.prog_scan_mode
        if self._scanmode == 1:
            self.camera_controller.exp_mode = "Edge Trigger"
            self.camera_controller.prog_scan_line_delay = self._scandelay
            self.camera_controller.exp_out_mode = 4
            print("camera ready to acquire programmable scan mode")

        else:
            self.camera_controller.exp_out_mode = "Any Row"
            self.camera_controller.exp_mode = "Edge Trigger"
            print("camera ready to acquire static light sheet mode")

        # prepare buffer pointer array
        # ptr_array = c_void_p * number_of_frames
        # data_ptr = ptr_array()
        # for i in range(number_of_frames):
        #     np_array = data_buffer[i]
        #     data_ptr[i] = np_array.ctypes.data

        self._numberofframes = number_of_frames
        self._data_buffer = data_buffer
        self._frames_received = 0
        self._frame_ids = []

        self.is_acquiring = True
        self.camera_controller.start_live(exp_time=self._exposuretime)

    def _receive_images(self):

        # wait_for_camera = True

        # while self.is_acquiring == True:

        # time.sleep(0.002)
        try:
            frame, fps, frame_count = self.camera_controller.poll_frame(
                timeout_ms=10000
            )
            # self._frame_ids.append(self._frames_received)

            self._data_buffer[self._frames_received][:, :] = np.copy(
                frame["pixel_data"][:]
            )
            frame = None
            del frame
            self._frames_received += 1
            if self._frames_received >= self._numberofframes:
                self._frames_received = 0
            return [self._frames_received - 1]
        except Exception as e:
            print(str(e))
        #     break

        # print("excited loop")

        return []

    # def _receive_imagesV1(self):
    #
    #     frame_ids = []
    #     wait_for_camera=True
    #     print("wait to receive frames")
    #
    #
    #     while wait_for_camera ==True:
    #         time.sleep(0.002)
    #         print(self.camera_controller.check_frame_status())
    #         if self.camera_controller.check_frame_status()=='FRAME_AVAILABLE':
    #             # framesReceived < number_of_frames:
    #
    #             if self.is_acquiring==False: #exit if acquisition is done
    #                 break
    #
    #             print(self._frames_received)
    #             try:
    #                 frame, fps, frame_count = self.camera_controller.poll_frame(
    #                 timeout_ms=10000)
    #                 print("received frame")
    #                 frame_ids.append(self._frames_received)
    #                 self.data_buffer[self._frames_received][:,:] = np.copy(frame[
    #                 'pixel_data'][:])
    #                 frame = None
    #                 del frame
    #                 self._frames_received += 1
    #                 if self._frames_received >= self._numberofframes:
    #                     self._frames_received = 0
    #
    #
    #
    #
    #             except Exception as e:
    #                 print(str(e))
    #                 break
    #     return frame_ids

    def close_image_series(self):
        """Close image series.

        Stops the acquisition and sets is_acquiring flag to False.
        """
        print("Calling finish")
        self.camera_controller.finish()
        self.is_acquiring = False

    def get_new_frame(self):
        """Get frame ids from Photometrics camera."""
        # return self.camera_controller.get_frame(exp_time=self._exposuretime)
        # self.camera_controller.start_live(exp_time=self._exposuretime)

        # self._receive_images()
        # return self._frame_ids[]

        return self._receive_images()

    def get_minimum_waiting_time(self):
        """Get minimum waiting time for HamamatsuOrca.

        This function get timing information from the camera device
        cyclic_trigger_period, minimum_trigger_blank, minimum_trigger_interval
        'cyclic_trigger_period' of current device is 0
        according to the document, trigger_blank should be bigger than trigger_interval.
        """
        # # cyclic_trigger = self.camera_controller.get_property_value(
        # 'cyclic_trigger_period')
        # trigger_blank = self.camera_controller.get_property_value(
        # 'minimum_trigger_blank')
        # # trigger_interval = self.camera_controller.get_property_value(
        # 'minimum_trigger_interval')
        # return trigger_blank
        # readout_time = (self.y_pixels + 1) * Camera_parameters.highres_
        # line_digitization_time  # +1 for the reset time at the first row
        # before the start
        # minimal_trigger_timeinterval = self._exposuretime + readout_time
        return 2



"""camera_settings.py

Note: Currently not used. Proof of concept to show how to change camera settings
in one file and have them applied to a camera from a different script.
"""

def apply_settings(camera):
    """Changes the settings of a camera."""
    camera.clear_mode = 0
    camera.exp_mode = "Internal Trigger"
    camera.readout_port = 0
    camera.speed_table_index = 0
    camera.gain = 1



# Copyright (c) 2021-2023  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

# https://iquilezles.org/articles/distfunctions/


def volume_from_sdf(sdf, N, pixel_size=1, subsample_z=1):
    """Generate an (N, N, N) image from sdf.

    Parameters
    ----------
    sdf : function
        A function that accepts (3, NxNxN) points as input and returns the Euclidean
        distance from each point to the object defined by the SDF.
    N : int
        Length of volume axis.
    pixel_size : float
        Rescale the image. Scaling must match object.
    subsample_z : int
        Return only the values every subsample slices.

    Returns
    -------
    npt.ArrayLike
        (N, N, N) image of SDF evaluated for each pixel.

    Examples
    --------
    >>> volume_from_sdf(lambda p: box(p, (15,15,30,)), 128)

    Generates a (128, 128, 128) image of a box sdf. The sdf is <= 0 within 128//2-30 to
    128//2+29 in z, 128//2-15 to 128//2+14 in x and y.
    """
    # Evaluate the SDF at the center of each pixel
    x = (np.arange(-N // 2, N // 2) + 0.5) * pixel_size
    z = (np.arange(-N // 2, N // 2, subsample_z) + 0.5) * pixel_size
    X, Y, Z = np.meshgrid(x, x, z)

    return sdf(np.vstack([X.ravel(), Y.ravel(), Z.ravel()])).reshape(N, N, -1).T


def sphere(p, R):
    """Signed distance function for a sphere.

    Parameters
    ----------
    p : npt.ArrayLike
        (3, N) array of points on which to evaluate the sdf.
    R : float
        Radius of the sphere.

    Returns
    -------
    npt.ArrayLike
        Signed distances from p to the sphere
    """
    return np.linalg.norm(p, axis=0) - R


def box(p, w):
    """Signed distance function for a box.

    Parameters
    ----------
    p : npt.ArrayLike
        (3, N) array of points on which to evaluate the sdf.
    w : tuple
        (3,) tuple of box half widths

    Returns
    -------
    npt.ArrayLike
        Signed distances from p to the box
    """
    w = np.array(w)  # Ensure w is an array

    q = np.abs(p) - w[:, None]
    return np.linalg.norm(np.maximum(q, 0.0), axis=0) + np.minimum(
        np.maximum(q[0, :], np.maximum(q[1, :], q[2, :])), 0.0
    )



# -*- coding: utf-8 -*-
"""
    This file refers to `ZhuangLab <https://github.com/ZhuangLab/storm-control>`,
    'dcamapi4.py' and 'dcam.py' This is a simplified version.

    Constants can be found at 'dcamsdk4/inc/dcamapi4.h' and 'dcamsdk4/inc/dcamprop.h'
    Function definitions can be found at 'dcamsdk4/doc/api_reference/dcamapi4_en.html'
"""

# Standard Library Imports
from ctypes import (
    windll,
    Structure,
    c_int32,
    c_void_p,
    sizeof,
    addressof,
    POINTER,
    byref,
    create_string_buffer,
    c_short,
    c_char_p,
    c_double,
)
from enum import IntEnum
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)

# ==== load shared library ====

__dll = windll.LoadLibrary("dcamapi.dll")

# ==== declare constants ====
DCAMBUF_ATTACHKIND_FRAME = 0
DCAMCAP_START_SEQUENCE = -1
DCAMWAIT_CAPEVENT_FRAMEREADY = 2
DCAMWAIT_CAPEVENT_STOPPED = 16
DCAMPROP_MODE__OFF = 1  # OFF
DCAMPROP_MODE__ON = 2  # ON


class DCAMPROP:
    """
    Valuable reference for understanding the values returned by the camera
    """

    class SENSORMODE(IntEnum):
        AREA = 1
        LINE = 3
        TDI = 4
        TDI_EXTENDED = 10
        PROGRESSIVE = 12
        SPLITVIEW = 14
        DUALLIGHTSHEET = 16
        PHOTONNUMBERRESOLVING = 18

    class SHUTTER_MODE(IntEnum):
        GLOBAL = 1
        ROLLING = 2

    class READOUTSPEED(IntEnum):
        SLOWEST = 1
        FASTEST = 0x7FFFFFFF

    class READOUT_DIRECTION(IntEnum):
        FORWARD = 1
        BACKWARD = 2
        BYTRIGGER = 3
        DIVERGE = 5
        FORWARDBIDIRECTION = 6
        REVERSEBIDIRECTION = 7

    class READOUT_UNIT(IntEnum):
        FRAME = 2
        BUNDLEDLINE = 3
        BUNDLEDFRAME = 4

    class CMOSMODE(IntEnum):
        NORMAL = 1
        NONDESTRUCTIVE = 2

    class OUTPUT_INTENSITY(IntEnum):
        NORMAL = 1
        TESTPATTERN = 2

    class OUTPUTDATA_OPERATION(IntEnum):
        RAW = 1
        ALIGNED = 2

    class TESTPATTERN_KIND(IntEnum):
        FLAT = 2
        IFLAT = 3
        HORZGRADATION = 4
        IHORZGRADATION = 5
        VERTGRADATION = 6
        IVERTGRADATION = 7
        LINE = 8
        ILINE = 9
        DIAGONAL = 10
        IDIAGONAL = 11
        FRAMECOUNT = 12

    class DIGITALBINNING_METHOD(IntEnum):
        MINIMUM = 1
        MAXIMUM = 2
        ODD = 3
        EVEN = 4
        SUM = 5
        AVERAGE = 6

    class TRIGGERSOURCE(IntEnum):
        INTERNAL = 1
        EXTERNAL = 2
        SOFTWARE = 3
        MASTERPULSE = 4

    class TRIGGERACTIVE(IntEnum):
        EDGE = 1
        LEVEL = 2
        SYNCREADOUT = 3
        POINT = 4

    class BUS_SPEED(IntEnum):
        SLOWEST = 1
        FASTEST = 0x7FFFFFFF

    class TRIGGER_MODE(IntEnum):
        NORMAL = 1
        PIV = 3
        START = 6

    class TRIGGERPOLARITY(IntEnum):
        NEGATIVE = 1
        POSITIVE = 2

    class TRIGGER_CONNECTOR(IntEnum):
        INTERFACE = 1
        BNC = 2
        MULTI = 3

    class INTERNALTRIGGER_HANDLING(IntEnum):
        SHORTEREXPOSURETIME = 1
        FASTERFRAMERATE = 2
        ABANDONWRONGFRAME = 3
        BURSTMODE = 4
        INDIVIDUALEXPOSURE = 7

    class SYNCREADOUT_SYSTEMBLANK(IntEnum):
        STANDARD = 1
        MINIMUM = 2

    class TRIGGERENABLE_ACTIVE(IntEnum):
        DENY = 1
        ALWAYS = 2
        LEVEL = 3
        START = 4

    class TRIGGERENABLE_POLARITY(IntEnum):
        NEGATIVE = 1
        POSITIVE = 2
        INTERLOCK = 3

    class OUTPUTTRIGGER_CHANNELSYNC(IntEnum):
        _1CHANNEL = 1
        _2CHANNELS = 2
        _3CHANNELS = 3

    class OUTPUTTRIGGER_PROGRAMABLESTART(IntEnum):
        FIRSTEXPOSURE = 1
        FIRSTREADOUT = 2

    class OUTPUTTRIGGER_SOURCE(IntEnum):
        EXPOSURE = 1
        READOUTEND = 2
        VSYNC = 3
        HSYNC = 4
        TRIGGER = 6

    class OUTPUTTRIGGER_POLARITY(IntEnum):
        NEGATIVE = 1
        POSITIVE = 2

    class OUTPUTTRIGGER_ACTIVE(IntEnum):
        EDGE = 1
        LEVEL = 2

    class OUTPUTTRIGGER_KIND(IntEnum):
        LOW = 1
        GLOBALEXPOSURE = 2
        PROGRAMABLE = 3
        TRIGGERREADY = 4
        HIGH = 5
        ANYROWEXPOSURE = 6

    class OUTPUTTRIGGER_BASESENSOR(IntEnum):
        VIEW1 = 1
        VIEW2 = 2
        ANYVIEW = 15
        ALLVIEWS = 16

    class EXPOSURETIME_CONTROL(IntEnum):
        OFF = 1
        NORMAL = 2

    class TRIGGER_FIRSTEXPOSURE(IntEnum):
        NEW = 1
        CURRENT = 2

    class TRIGGER_GLOBALEXPOSURE(IntEnum):
        NONE = 1
        ALWAYS = 2
        DELAYED = 3
        EMULATE = 4
        GLOBALRESET = 5

    class FIRSTTRIGGER_BEHAVIOR(IntEnum):
        STARTEXPOSURE = 1
        STARTREADOUT = 2

    class MASTERPULSE_MODE(IntEnum):
        CONTINUOUS = 1
        START = 2
        BURST = 3

    class MASTERPULSE_TRIGGERSOURCE(IntEnum):
        EXTERNAL = 1
        SOFTWARE = 2

    class MECHANICALSHUTTER(IntEnum):
        AUTO = 1
        CLOSE = 2
        OPEN = 3

    class LIGHTMODE(IntEnum):
        LOWLIGHT = 1
        HIGHLIGHT = 2

    class SENSITIVITYMODE(IntEnum):
        OFF = 1
        ON = 2
        INTERLOCK = 3

    class EMGAINWARNING_STATUS(IntEnum):
        NORMAL = 1
        WARNING = 2
        PROTECTED = 3

    class PHOTONIMAGINGMODE(IntEnum):
        _0 = 0
        _1 = 1
        _2 = 2
        _3 = 3

    class SENSORCOOLER(IntEnum):
        OFF = 1
        ON = 2
        MAX = 4

    class SENSORTEMPERATURE_STATUS(IntEnum):
        NORMAL = 0
        WARNING = 1
        PROTECTION = 2

    class SENSORCOOLERSTATUS(IntEnum):
        ERROR4 = -4
        ERROR3 = -3
        ERROR2 = -2
        ERROR1 = -1
        NONE = 0
        OFF = 1
        READY = 2
        BUSY = 3
        ALWAYS = 4
        WARNING = 5

    class REALTIMEGAINCORRECT_LEVEL(IntEnum):
        _1 = 1
        _2 = 2
        _3 = 3
        _4 = 4
        _5 = 5

    class WHITEBALANCEMODE(IntEnum):
        FLAT = 1
        AUTO = 2
        TEMPERATURE = 3
        USERPRESET = 4

    class DARKCALIB_TARGET(IntEnum):
        ALL = 1
        ANALOG = 2

    class SHADINGCALIB_METHOD(IntEnum):
        AVERAGE = 1
        MAXIMUM = 2
        USETARGET = 3

    class CAPTUREMODE(IntEnum):
        NORMAL = 1
        DARKCALIB = 2
        SHADINGCALIB = 3
        TAPGAINCALIB = 4
        BACKFOCUSCALIB = 5

    class INTERFRAMEALU_ENABLE(IntEnum):
        OFF = 1
        TRIGGERSOURCE_ALL = 2
        TRIGGERSOURCE_INTERNAL = 3

    class SHADINGCALIB_DATASTATUS(IntEnum):
        NONE = 1
        FORWARD = 2
        BACKWARD = 3
        BOTH = 4

    class TAPGAINCALIB_METHOD(IntEnum):
        AVE = 1
        MAX = 2
        MIN = 3

    class RECURSIVEFILTERFRAMES(IntEnum):
        _2 = 2
        _4 = 4
        _8 = 8
        _16 = 16
        _32 = 32
        _64 = 64

    class INTENSITYLUT_MODE(IntEnum):
        THROUGH = 1
        PAGE = 2
        CLIP = 3

    class BINNING(IntEnum):
        _1 = 1
        _2 = 2
        _4 = 4
        _8 = 8
        _16 = 16
        _1_2 = 102
        _2_4 = 204

    class COLORTYPE(IntEnum):
        BW = 0x00000001
        RGB = 0x00000002
        BGR = 0x00000003

    class BITSPERCHANNEL(IntEnum):
        _8 = 8
        _10 = 10
        _12 = 12
        _14 = 14
        _16 = 16

    class DEFECTCORRECT_MODE(IntEnum):
        OFF = 1
        ON = 2

    class DEFECTCORRECT_METHOD(IntEnum):
        CEILING = 3
        PREVIOUS = 4

    class HOTPIXELCORRECT_LEVEL(IntEnum):
        STANDARD = 1
        MINIMUM = 2
        AGGRESSIVE = 3

    class DEVICEBUFFER_MODE(IntEnum):
        THRU = 1
        SNAPSHOT = 2

    class SYSTEM_ALIVE(IntEnum):
        OFFLINE = 1
        ONLINE = 2
        ERROR = 3

    class TIMESTAMP_MODE(IntEnum):
        NONE = 1
        LINEBEFORELEFT = 2
        LINEOVERWRITELEFT = 3
        AREABEFORELEFT = 4
        AREAOVERWRITELEFT = 5

    class TIMING_EXPOSURE(IntEnum):
        AFTERREADOUT = 1
        OVERLAPREADOUT = 2
        ROLLING = 3
        ALWAYS = 4
        TDI = 5

    class TIMESTAMP_PRODUCER(IntEnum):
        NONE = 1
        DCAMMODULE = 2
        KERNELDRIVER = 3
        CAPTUREDEVICE = 4
        IMAGINGDEVICE = 5

    class FRAMESTAMP_PRODUCER(IntEnum):
        NONE = 1
        DCAMMODULE = 2
        KERNELDRIVER = 3
        CAPTUREDEVICE = 4
        IMAGINGDEVICE = 5

    class CAMERASTATUS_INTENSITY(IntEnum):
        GOOD = 1
        TOODARK = 2
        TOOBRIGHT = 3
        UNCARE = 4
        EMGAIN_PROTECTION = 5
        INCONSISTENT_OPTICS = 6
        NODATA = 7

    class CAMERASTATUS_INPUTTRIGGER(IntEnum):
        GOOD = 1
        NONE = 2
        TOOFREQUENT = 3

    class CAMERASTATUS_CALIBRATION(IntEnum):
        DONE = 1
        NOTYET = 2
        NOTRIGGER = 3
        TOOFREQUENTTRIGGER = 4
        OUTOFADJUSTABLERANGE = 5
        UNSUITABLETABLE = 6
        TOODARK = 7
        TOOBRIGHT = 8
        NOTDETECTOBJECT = 9

    class CONFOCAL_SCANMODE(IntEnum):
        SIMULTANEOUS = 1
        SEQUENTIAL = 2

    class SUBUNIT_CONTROL(IntEnum):
        NOTINSTALLED = 0
        OFF = 1
        ON = 2

    class SUBUNIT_PINHOLESIZE(IntEnum):
        ERROR = 1
        SMALL = 2
        MEDIUM = 3
        LARGE = 4

    class MODE(IntEnum):
        OFF = 1
        ON = 2


class DCAMDEV_OPEN(Structure):
    _pack_ = 8
    _fields_ = [("size", c_int32), ("index", c_int32), ("hdcam", c_void_p)]  # out

    def __init__(self):
        self.size = sizeof(DCAMDEV_OPEN)
        self.index = 0


class DCAMAPI_INIT(Structure):
    _pack_ = 8
    _fields_ = [
        ("size", c_int32),
        ("iDeviceCount", c_int32),  # out
        ("reserved", c_int32),
        ("initoptionbytes", c_int32),
        ("initoption", POINTER(c_int32)),
        ("guid", c_void_p),  # const DCAM_GUID*
    ]

    def __init__(self):
        self.size = sizeof(DCAMAPI_INIT)


class DCAMWAIT_OPEN(Structure):
    _pack_ = 8
    _fields_ = [
        ("size", c_int32),
        ("supportevent", c_int32),  # out
        ("hwait", c_void_p),  # out
        ("hdcam", c_void_p),
    ]

    def __init__(self):
        self.size = sizeof(DCAMWAIT_OPEN)


class DCAMWAIT_START(Structure):
    _pack_ = 8
    _fields_ = [
        ("size", c_int32),
        ("eventhappened", c_int32),  # out
        ("eventmask", c_int32),
        ("timeout", c_int32),
    ]

    def __init__(self):
        self.size = sizeof(DCAMWAIT_START)


class DCAMERR(IntEnum):
    # success
    SUCCESS = (
        1  # 1, no error, general success code, app should check the value is positive
    )
    ALREADYOPENED = -520093694  # 0xE1000002
    NOTSUPPORT = (
        -2147479805
    )  # 0x80000f03, camera does not support the function or property
    # with current settings
    TIMEOUT = -2147483386  # 0x80000106, timeout

    # status error
    BUSY = -2147483391  # 0x80000101, API cannot process in busy state.
    NOTREADY = -2147483389  # 0x80000103, API requires ready state.
    NOTSTABLE = -2147483388  # 0x80000104, API requires stable or unstable state.
    UNSTABLE = -2147483387  # 0x80000105, API does not support in unstable state.

    # calling error
    INVALIDCAMERA = -2147481594  # 0x80000806, invalid camera
    INVALIDHANDLE = -2147481593  # 0x80000807, invalid camera handle
    INVALIDPARAM = -2147481592  # 0x80000808, invalid parameter
    INVALIDVALUE = -2147481567  # 0x80000821, invalid property value
    OUTOFRANGE = -2147481566  # 0x80000822, value is out of range
    NOTWRITABLE = -2147481565  # 0x80000823, the property is not writable
    NOTREADABLE = -2147481564  # 0x80000824, the property is not readable
    INVALIDPROPERTYID = -2147481563  # 0x80000825, the property id is invalid


class DCAMPROP_ATTR(Structure):
    _pack_ = 8
    _fields_ = [
        ("cbSize", c_int32),
        ("iProp", c_int32),
        ("option", c_int32),
        ("iReserved1", c_int32),
        ("attribute", c_int32),
        ("iGroup", c_int32),
        ("iUnit", c_int32),
        ("attribute2", c_int32),
        ("valuemin", c_double),
        ("valuemax", c_double),
        ("valuestep", c_double),
        ("valuedefault", c_double),
        ("nMaxChannel", c_int32),
        ("iReserved3", c_int32),
        ("nMaxView", c_int32),
        ("iProp_NumberOfElement", c_int32),
        ("iProp_ArrayBase", c_int32),
        ("iPropStep_Element", c_int32),
    ]

    def __init__(self):
        self.cbSize = sizeof(DCAMPROP_ATTR)


class DCAMBUF_ATTACH(Structure):
    _pack_ = 8
    _fields_ = [
        ("size", c_int32),  # sizeof(*this)
        ("iKind", c_int32),  # DCAMBUF_ATTAHKIND: DCAMBUF_ATTACHKIND_FRAME = 0
        ("buffer", POINTER(c_void_p)),  # array of buffer pointers
        ("buffercount", c_int32),  # number of pointers in array "buffer"
    ]

    def __init__(self):
        self.size = sizeof(DCAMBUF_ATTACH)


class DCAMCAP_TRANSFERINFO(Structure):
    _pack_ = 8
    _fields_ = [
        ("size", c_int32),
        ("iKind", c_int32),
        ("nNewestFrameIndex", c_int32),
        ("nFrameCount", c_int32),
    ]

    def __init__(self):
        self.size = sizeof(DCAMCAP_TRANSFERINFO)
        self.iKind = 0
        self.nNewestFrameIndex = -1
        self.nFrameCount = 0


class DCAMDATA_HDR(Structure):
    _pack_ = 8
    _fields_ = [
        ("size", c_int32),
        ("iKind", c_int32),
        ("option", c_int32),
        ("reserved2", c_int32),
    ]

    def __init__(self):
        self.size = sizeof(DCAMDATA_HDR)
        self.iKind = 0
        self.option = 0
        self.reserved2 = 0


class DCAMDATA_REGIONRECT(Structure):
    _pack_ = 8
    _fields_ = [
        ("left", c_short),
        ("top", c_short),
        ("right", c_short),
        ("bottom", c_short),
    ]


class DCAMDATA_REGION(Structure):
    _pack_ = 8
    _fields_ = [
        ("hdr", POINTER(DCAMDATA_HDR)),
        ("option", c_int32),
        ("type", c_int32),
        ("data", c_void_p),
        ("datasize", c_int32),
        ("reserved", c_int32),
    ]


class DCAMDEV_STRING(Structure):
    _fields_ = [
        ("size", c_int32),
        ("iString", c_int32),
        ("text", c_char_p),
        ("textbytes", c_int32),
    ]


property_dict = {
    "exposure_time": 2031888,  # 0x001F0110, R/W, sec, "EXPOSURE TIME"
    "sensor_mode": 4194832,  # 0x00400210, R/W, mode,  "SENSOR MODE"
    "defect_correct_mode": 4653072,  # 0x00470010, R/W, mode,  "DEFECT CORRECT MODE"
    "binning": 4198672,  # 0x00401110, R/W, mode, "BINNING"
    "readout_speed": 4194576,  # 0x00400110, R/W, long,    "READOUT SPEED"
    "readout_direction": 4194608,  # 0x00400130, R/W, mode,   "READOUT DIRECTION"
    "readout_time": 4206608,  # 0x00403010, R/O, sec,   "TIMING READOUT TIME"
    "trigger_active": 1048864,  # 0x00100120, R/W, mode,   "TRIGGER ACTIVE"
    "trigger_mode": 1049104,  # 0x00100210, R/W, mode,    "TRIGGER MODE"
    "trigger_polarity": 1049120,  # 0x00100220, R/W, mode, "TRIGGER POLARITY"
    "trigger_source": 1048848,  # 0x00100110, R/W, mode,   "TRIGGER SOURCE"
    "trigger_delay": 1049184,  # 0x00100260,	/* R/W, sec,	"TRIGGER DELAY"
    "internal_line_interval": 4208720,  # 0x00403850, R/W, sec,
    # "INTERNAL LINE INTERVAL"
    "image_width": 4325904,  # 0x00420210, R/O, long, "IMAGE WIDTH"
    "image_height": 4325920,  # 0x00420220, R/O, long,    "IMAGE HEIGHT"
    "exposuretime_control": 2031920,  # 0x001F0130, R/W, mode,
    # "EXPOSURE TIME CONTROL"
    "subarray_hpos": 4202768,  # 0x00402110, R/W, long,    "SUBARRAY HPOS"
    "subarray_hsize": 4202784,  # 0x00402120, R/W, long,   "SUBARRAY HSIZE"
    "subarray_vpos": 4202800,  # 0x00402130, R/W, long,    "SUBARRAY VPOS"
    "subarray_vsize": 4202816,  # 0x00402140, R/W, long,   "SUBARRAY VSIZE"
    "subarray_mode": 4202832,  # 0x00402150, R/W, mode,    "SUBARRAY MODE"
    "cyclic_trigger_period": 4206624,  # 0x00403020, R/O, sec,
    # "TIMING CYCLIC TRIGGER PERIOD"
    "minimum_trigger_blank": 4206640,  # 0x00403030, R/O, sec,
    # "TIMING MINIMUM TRIGGER BLANKING"
    "minimum_trigger_interval": 4206672,  # 0x00403050, R/O, sec,
    # "TIMING MINIMUM TRIGGER INTERVAL"
    "pixel_width": 4327440,  # 0x00420810, R/O, micro-meter,
    # "IMAGE DETECTOR PIXEL WIDTH"
    "pixel_height": 4327456,  # 0x00420820, R/O, micro-meter,
    # "IMAGE DETECTOR PIXEL HEIGHT"
}


# ==== api function references ====
dcamapi_init = __dll.dcamapi_init
dcamapi_uninit = __dll.dcamapi_uninit
dcamdev_open = __dll.dcamdev_open
dcamdev_close = __dll.dcamdev_close
dcamwait_open = __dll.dcamwait_open
dcamwait_close = __dll.dcamwait_close
dcamprop_getattr = __dll.dcamprop_getattr
dcamprop_getvalue = __dll.dcamprop_getvalue
dcamprop_setvalue = __dll.dcamprop_setvalue
dcamprop_setgetvalue = __dll.dcamprop_setgetvalue
dcambuf_attach = __dll.dcambuf_attach
dcambuf_release = __dll.dcambuf_release
dcamcap_start = __dll.dcamcap_start
dcamcap_stop = __dll.dcamcap_stop
dcamcap_transferinfo = __dll.dcamcap_transferinfo
dcamwait_start = __dll.dcamwait_start
dcamwait_abort = __dll.dcamwait_abort
dcamcap_firetrigger = __dll.dcamcap_firetrigger
# dcamdev_setdata = __dll.dcamdev_setdata
dcamdev_getstring = __dll.dcamdev_getstring


class camReg(object):
    """
    Keep track of the number of cameras initialised so we can initialise and
    finalise the library.

    Cribbed from https://github.com/python-microscopy/python-microscopy/
                 blob/master/PYME/Acquire/Hardware/HamamatsuDCAM/HamamatsuDCAM.py
    """

    numCameras = 0
    maxCameras = 0

    @classmethod
    def regCamera(cls):
        if cls.numCameras <= 0:
            # Initialize the API
            paraminit = DCAMAPI_INIT()
            if int(dcamapi_init(byref(paraminit))) < 0:
                # NOTE: This is an AttributeError to match the other error thrown by
                # this class in startup functions. This really makes no sense as an
                # attribute error.
                dcamapi_uninit()
                raise Exception("DCAM initialization failed.")
            cls.maxCameras = paraminit.iDeviceCount
            cls.numCameras = 0

        cls.numCameras += 1
        # print(f"Number of cameras is {cls.numCameras}")

    @classmethod
    def unregCamera(cls):
        cls.numCameras -= 1
        if cls.numCameras == 0:
            dcamapi_uninit()


class DCAM:
    def __init__(self, index=0):
        self.__hdcam = 0
        self.__hdcamwait = 0

        # open camera
        self.dev_open(index)
        self.__open_hdcamwait()

        # TODO: Is turning binning off necessary?
        self.prop_setvalue(property_dict["subarray_mode"], DCAMPROP_MODE__OFF)
        _, self.max_image_width, _ = self.get_property_range("image_width")
        _, self.max_image_height, _ = self.get_property_range("image_height")
        self.min_image_width, _, self.step_image_width = self.get_property_range(
            "subarray_hsize"
        )
        self.min_image_height, _, self.step_image_height = self.get_property_range(
            "subarray_vsize"
        )

        # print(f"steps {self.step_image_height} {self.step_image_width}")

        self._serial_number = self.get_string_value(
            c_int32(int("0x04000102", 0))
        ).strip("S/N: ")

    def __del__(self):
        self.dev_close()

    def __result(self, errvalue):
        """
        Internal use. Keep last error code
        """
        if errvalue < 0:
            try:
                logger.debug(f"error message: {DCAMERR(errvalue)}")
            except Exception:
                logger.debug(f"error message: {errvalue}")
            return False

        return True

    # dcamwait functions

    def __open_hdcamwait(self):
        """
        Get HDCAMWAIT handle
        """
        if not self.__hdcamwait == 0:
            return True

        paramwaitopen = DCAMWAIT_OPEN()
        paramwaitopen.hdcam = self.__hdcam
        ret = self.__result(dcamwait_open(byref(paramwaitopen)))
        if ret is False:
            return False

        if paramwaitopen.hwait == 0:
            return self.__result(DCAMERR.INVALIDWAITHANDLE)

        self.__hdcamwait = c_void_p(paramwaitopen.hwait)
        return True

    def __close_hdcamwait(self):
        """
        Close HDCAMWAIT handle
        """

        if self.__hdcamwait == 0:
            return True

        ret = self.__result(dcamwait_close(self.__hdcamwait))
        if ret is False:
            return False

        self.__hdcamwait = 0
        return True

    def dev_open(self, index=-1):
        """
        Get HDCAM handle for controling camera.
        After calling close(), call this again if you need to resume measurement.

        Args:
            arg1(int): device index

        Returns:
            True:   if dcamdev_open() succeeded.
            False:  if dcamdev_open() returned DCAMERR except SUCCESS.  lasterr()
                    returns the DCAMERR value.
        """
        # print("Opening Camera")
        if self.__hdcam:
            # print("Camera already open")
            return self.__result(
                DCAMERR.ALREADYOPENED
            )  # instance is already opened. New Error.

        camReg.regCamera()  # Make sure DCAMAPI is initialized

        paramopen = DCAMDEV_OPEN()
        if index >= 0:
            paramopen.index = index
        else:
            paramopen.index = self.__iDevice

        ret = self.__result(dcamdev_open(byref(paramopen)))
        if ret is False:
            print("Camera dev_open failed")
            return False

        self.__hdcam = c_void_p(paramopen.hdcam)

        return True

    def dev_close(self):
        """
        Close dcam handle.
        Call this if you need to close the current device.

        Returns:
            True:
        """
        if self.__hdcam:
            self.__close_hdcamwait()
            dcamdev_close(self.__hdcam)
            self.__hdcam = 0

        camReg.unregCamera()

        return True

    # dcamprop functions

    def prop_getattr(self, idprop):
        """
        Get property attribute

        args:
            arg1(DCAM_IDPROP): property id

        Returns:
            DCAMPROP_ATTR
            if False, error happened.  lasterr() returns the DCAMERR value
        """
        if not self.__hdcam:
            self.__result(DCAMERR.INVALIDHANDLE)  # instance is not opened yet.
            return None

        propattr = DCAMPROP_ATTR()
        propattr.iProp = idprop
        ret = self.__result(dcamprop_getattr(self.__hdcam, byref(propattr)))
        if ret is False:
            return None

        return propattr

    def prop_getvalue(self, idprop):
        """
        Get property value

        args:
            arg1(DCAM_IDPROP): property id

        Returns:
            double
            if False, error happened.  lasterr() returns the DCAMERR value
        """
        if not self.__hdcam:
            return self.__result(DCAMERR.INVALIDHANDLE)  # instance is not opened yet.

        cDouble = c_double()
        ret = self.__result(dcamprop_getvalue(self.__hdcam, idprop, byref(cDouble)))
        # print("prop_getvalue response:", cDouble.value)
        # print("property id:", idprop)
        if ret is False:
            return False

        return cDouble.value

    def prop_setvalue(self, idprop, fValue):
        """
        Set property value

        args:
            arg1(DCAM_IDPROP): property id
            arg2(double): setting value

        Returns:
            True   success
            False  error happened.  lasterr() returns the DCAMERR value
        """
        if not self.__hdcam:
            return self.__result(DCAMERR.INVALIDHANDLE)  # instance is not opened yet.

        fValue = c_double(fValue)
        ret = self.__result(dcamprop_setvalue(self.__hdcam, idprop, fValue))
        if ret is False:
            return False

        return True

    def prop_setgetvalue(self, idprop, fValue, option=0):
        """
        Set and get property value

        args:
            arg1(DCAM_IDPROP): property id
            arg2(double): input value for setting and receive actual set value by ref

        Returns:
            double
            if False, error happened.  lasterr() returns the DCAMERR value
        """
        if not self.__hdcam:
            return self.__result(DCAMERR.INVALIDHANDLE)  # instance is not opened yet.

        cDouble = c_double(fValue)
        cOption = c_int32(option)
        cIDProp = c_int32(idprop)
        ret = self.__result(
            dcamprop_setgetvalue(self.__hdcam, cIDProp, byref(cDouble), cOption)
        )
        if ret is False:
            return False

        return cDouble.value

    def get_property_range(self, name):
        """
        # Returns the range of appropriate values start, stop, step
        """
        if name not in property_dict:
            print(
                "could not set value for",
                name,
                "please make sure the property name",
                "is correct and is added to property_dict!",
            )
            return [None, None, None]

        # get property code setPropertyValue
        idprop = property_dict[name]

        property_attribute = self.prop_getattr(idprop)
        if property_attribute is not None:
            return [
                float(property_attribute.valuemin),
                float(property_attribute.valuemax),
                float(property_attribute.valuestep),
            ]
        return [None, None, None]

    def set_property_value(self, name, value):
        """
        # this function will set property value according to property name
        """
        if name not in property_dict:
            print(
                "could not set value for",
                name,
                "please make sure the property name",
                "is correct and is added to property_dict!",
            )
            return False
        # get property code setPropertyValue
        idprop = property_dict[name]

        # Find property limits and correct value if necessary
        (
            property_value_min,
            property_value_max,
            property_value_step,
        ) = self.get_property_range(name)

        if property_value_min is None:
            print("Could not set attribute", name)
            print("property range:", name, property_value_min, property_value_max)
            return False

        # Cast value to a multiple of step size
        if property_value_step > 0:
            value = int(value // property_value_step) * property_value_step

        if value < property_value_min:
            print(
                " The property value of ",
                value,
                "is less than minimum of",
                property_value_min,
                name,
                "setting to minimum",
            )
            value = property_value_min
        if value > property_value_max:
            print(
                " The property value of",
                value,
                "is greater than maximum of",
                property_value_max,
                name,
                "setting to maximum",
            )
            value = property_value_max

        # Set value and get what is set
        final_configuration = self.prop_setgetvalue(idprop, value)

        if abs(final_configuration - value) < value / 100:
            return True
        else:
            print(name, "Configuration Failed", value, final_configuration)
            return False

    def get_property_value(self, name):
        """
        Provides the idprop value after looking it up in the property_dict
        """
        return self.prop_getvalue(property_dict[name])

    def get_string_value(self, id_str):
        """
        Get attribute value from the camera by IDSTR (or ERR) as a string.

        Parameters
        ----------
        id_str : c_int32

        Returns
        -------
        Value of IDSTR or ERR as a string.
        """

        c_buf_len = 256
        c_buf = create_string_buffer(c_buf_len)
        param = DCAMDEV_STRING()
        param.size = sizeof(param)
        param.text = addressof(c_buf)
        param.textbytes = c_int32(c_buf_len)
        param.iString = id_str
        dcamdev_getstring(self.__hdcam, byref(param))

        return c_buf.value.decode()

    def set_ROI(self, left, top, right, bottom):
        """
        # this function set 'subarray' properties.
        # Width - Can be an even or odd number.
        # Height - Can be an even number.
        # Horizontal Position - Can be even or odd number.
        # Vertical Position - Can be even number.
        # Bottom must be even number.
        """
        # Set subarray mode off. This setting is not mandatory, but you have to control
        # the setting order of offset and size when mode is on
        self.prop_setgetvalue(property_dict["subarray_mode"], DCAMPROP_MODE__OFF)

        left = int(left // self.step_image_width) * self.step_image_width
        top = int(top // self.step_image_height) * self.step_image_height
        hsize = int((right - left) // self.step_image_width) * self.step_image_width
        vsize = int((bottom - top) // self.step_image_height) * self.step_image_height

        self.prop_setvalue(property_dict["subarray_hpos"], left)
        self.prop_setvalue(property_dict["subarray_vpos"], top)
        self.prop_setvalue(property_dict["subarray_hsize"], hsize)
        self.prop_setvalue(property_dict["subarray_vsize"], vsize)

        # set subarray on
        self.prop_setgetvalue(property_dict["subarray_mode"], DCAMPROP_MODE__ON)

        # Update minimum image sizes based on subarray mode
        self.min_image_width, _, self.step_image_width = self.get_property_range(
            "subarray_hsize"
        )
        self.min_image_height, _, self.step_image_height = self.get_property_range(
            "subarray_vsize"
        )

        return (
            self.prop_getvalue(property_dict["image_width"]),
            self.prop_getvalue(property_dict["image_height"]),
        )

    def start_acquisition(self, data_buffer, number_of_frames=100):
        """
        # this function will initialize parameters, attach buffer, start capture
        """
        # initialize buffer index
        self.pre_index = -1
        self.number_of_frames = number_of_frames

        # prepare buffer pointer array
        self.data_buffer = data_buffer
        ptr_array = c_void_p * number_of_frames
        self.data_ptr = ptr_array()
        for i in range(number_of_frames):
            np_array = data_buffer[i]
            self.data_ptr[i] = np_array.ctypes.data

        # attach buffer
        attach_param = DCAMBUF_ATTACH()
        attach_param.iKind = DCAMBUF_ATTACHKIND_FRAME
        attach_param.buffer = self.data_ptr
        attach_param.buffercount = self.number_of_frames
        if self.__result(dcambuf_attach(self.__hdcam, attach_param)):
            self.pre_frame_count = 0
            self.pre_index = 0
            return self.__result(dcamcap_start(self.__hdcam, DCAMCAP_START_SEQUENCE))
        return False

    def stop_acquisition(self):
        """
        # this function will stop capture and detach buffer
        """
        # stop capture
        dcamcap_stop(self.__hdcam)

        # detach buffer
        return self.__result(dcambuf_release(self.__hdcam, DCAMBUF_ATTACHKIND_FRAME))

    def get_frames(self):
        """
        # this function will return a list of frame index
        # following lines are from documents: 'dcamapi4-en.html'
        # The host software can check the capturing status with the dcamcap_status()
        # function. It will return a DCAMCAP_STATUS.
        # The dcamcap_transferinfo() function returns the total number of images
        # captured and the frame index of the last captured image.
        # These functions do not wait for any events and will return with their
        # values immediately.
        # These functions are useful for polling however this is stressful to the CPU.
        # We recommend using dcamwait functions to wait for events such as the arrival
        # of new frames, then calling these functions to check status and/or
        # transferred information.
        """
        # current solution: wait for a new frame, then call dcamcap_transferinfo()
        frame_idx_list = []
        wait_param = DCAMWAIT_START()
        wait_param.eventmask = DCAMWAIT_CAPEVENT_FRAMEREADY | DCAMWAIT_CAPEVENT_STOPPED
        wait_param.timeout = 500  # 500ms
        #  Timeout Duration - Will throw an error if the timeout is too small.
        #  Currently set to a value > maximum typical integration time.

        if self.__result(dcamwait_start(self.__hdcamwait, byref(wait_param))):
            cap_info = DCAMCAP_TRANSFERINFO()
            dcamcap_transferinfo(self.__hdcam, byref(cap_info))
            # after testing with the camera, we find out that:
            # nNewestFrameIndex starts from 0;
            # nFrameCount increments all the frames from beginning even we have already
            # pulled the info out
            frame_count = cap_info.nFrameCount - self.pre_frame_count

            # print("Frame Count - cap_info", cap_info.nFrameCount, frame_count)
            # print("Newest Frame Index - cap_info", cap_info.nNewestFrameIndex)

            if frame_count <= cap_info.nNewestFrameIndex + 1:
                frame_idx_list = list(
                    range(
                        cap_info.nNewestFrameIndex - frame_count + 1,
                        cap_info.nNewestFrameIndex + 1,
                    )
                )
            else:
                if frame_count > self.number_of_frames:
                    # We waited so long that we've missed at least a buffer's
                    # worth of frames
                    msg = (
                        "Hamamatsu API - Dropping frames! Number of frames "
                        f"since last poll {frame_count} exceeds buffer size  "
                        f"{self.number_of_frames}. This is likely due to slow "
                        "file saving."
                    )
                    logger.debug(msg)
                    print(msg)

                    # Update the frame count to grab the last N usable frames
                    frame_count = self.number_of_frames

                frame_idx_list = list(
                    range(
                        self.number_of_frames
                        - frame_count
                        + cap_info.nNewestFrameIndex
                        + 1,
                        self.number_of_frames,
                    )
                ) + list(range(0, cap_info.nNewestFrameIndex + 1))

            # check if backlog happens
            # if (self.pre_index+1) % self.number_of_frames != frame_idx_list[0]:
            #    print('backlog happens!')
            self.pre_index = cap_info.nNewestFrameIndex
            self.pre_frame_count = cap_info.nFrameCount

        return frame_idx_list

    def get_camera_handler(self):
        return self.__hdcam

    def fire_software_trigger(self):
        trigger_source = self.get_property_value("trigger_source")
        if trigger_source == 3.0:
            # fire trigger to camera
            err = dcamcap_firetrigger(self.__hdcam, 0)
            if err < 0:
                print("an error happened when sending trigger to the camera", err)
        else:
            print(f"Camera is in mode {trigger_source}, not software mode (3).")


if __name__ == "__main__":
    print("start testing Hamamatsu API!")

    # create shared memory buffer
    import sys

    sys.path.append("../../../concurrency")
    from concurrency_tools import SharedNDArray
    import threading
    import time

    number_of_frames = 200
    # TODO: Get the shape of the image from the configuration file (at least for the
    # real software not testing)
    # CameraParameters:
    # # Hamamatsu Orca Flash 4.0
    # number_of_cameras: 1
    # x_pixels: 2048.0
    # y_pixels: 2048.0

    start_time = time.time()
    data_buffer = [
        SharedNDArray(shape=(2048, 2048), dtype="uint16")
        for i in range(number_of_frames)
    ]
    stop_time = time.time()
    print(
        "Duration of time to create a buffer for ",
        number_of_frames,
        stop_time - start_time,
    )
    # Grows linearly with the number of frames.

    # start camera
    camera = DCAM(0)

    # initialize camera
    configuration = {
        "image_width": 2048.0,
        "image_height": 2048.0,
        "sensor_mode": 12,  # 12 for progressive
        "defect_correct_mode": 2.0,
        "binning": 1.0,
        "readout_speed": 1.0,
        "trigger_active": 1.0,
        "trigger_mode": 1.0,  # external light-sheet mode
        "trigger_polarity": 2.0,  # positive pulse
        "trigger_source": 3.0,  # software
        "exposure_time": 0.02,
        "internal_line_interval": 0.000075,
    }
    # camera.prop_getvalue(property_dict['exposuretime_control'])

    # configure camera
    for key in configuration:
        # INVALIDVALUE = -2147481567  # 0x80000821, invalid property value
        # INVALIDPROPERTYID = -2147481563  # 0x80000825, the property id is invalid
        # print("property:", property_dict[key], " key:", key)
        camera.prop_setvalue(property_dict[key], configuration[key])

    # start data process
    def data_func():
        while True:
            frames = camera.get_frames()
            if not frames:
                break
            # print('get image frame:', frames)

    def test_acquisition():
        data_process = threading.Thread(target=data_func)
        data_process.name = "HamamatsuAPI Data Process"
        data_process.start()

        # start acquisition
        camera.start_acquisition(data_buffer, number_of_frames)

        # set camera that trigger from software
        TRIGGERSOURCE_SOFTWARE = 3
        if camera.prop_setgetvalue(
            property_dict["trigger_source"], TRIGGERSOURCE_SOFTWARE
        ):

            # fire trigger to camera
            for i in range(10):
                err = dcamcap_firetrigger(camera.get_camera_handler(), 0)
                if err < 0:
                    print("an error happened when sending trigger to the camera", err)
                    break
                time.sleep(configuration["exposure_time"] + 0.005)
        # end acquisition
        camera.stop_acquisition()
        data_process.join()

    def get_buffer_duration():
        # get time cost of attaching and detaching buffer
        number_of_frames = 0
        for i in range(20):
            number_of_frames += 100
            start_time = time.perf_counter()
            camera.start_acquisition()
            camera.stop_acquisition()
            end_time = time.perf_counter()
            print(
                "time cost to attach a buffer(",
                number_of_frames,
                "):",
                end_time - start_time,
            )

    # test ROI setting
    def test_ROI(roi_height=2048, roi_width=2048):
        camera_height = 2048
        camera_width = 2048

        roi_top = (camera_height - roi_height) / 2
        roi_bottom = roi_top + roi_height - 1
        roi_left = (camera_width - roi_width) / 2
        roi_right = roi_left + roi_width - 1

        width, height = camera.set_ROI(roi_left, roi_top, roi_right, roi_bottom)

        print("image width and height:", width, height)
        print("subarray_hpos", camera.prop_getvalue(property_dict["subarray_hpos"]))
        print("subarray_hsize", camera.prop_getvalue(property_dict["subarray_hsize"]))
        print("subarray_vpos", camera.prop_getvalue(property_dict["subarray_vpos"]))
        print("subarray_vsize", camera.prop_getvalue(property_dict["subarray_vsize"]))
        print(
            "sub array mode(1: OFF, 2: ON): ",
            camera.prop_getvalue(property_dict["subarray_mode"]),
        )

    test_ROI(512, 512)

    data_buffer = [
        SharedNDArray(shape=(512, 512), dtype="uint16") for i in range(number_of_frames)
    ]
    test_acquisition()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Imports

# Third Party Imports
import numpy as np

# Local imports


def affine_rotation(x=0, y=0, z=0):
    """Calculate the general 3D affine transform for rotation.

    Returns a rotation matrix for a rotation about the x, y, and z axes.

    Rotation about x:
    [1, 0, 0, 0,
    0, cos(theta), -sin(theta), 0,
    0, sin(theta), cos(theta), 0,
    0, 0, 0, 1]

    Rotation about Y:
    [cos(theta), 0, sin(theta), 0,
    0, 1, 0, 0,
    -sin(theta), 0, cos(theta), 0,
    0, 0, 0, 1]

    Rotation about Z:
    [cos(theta), -sin(theta), 0, 0,
    sin(theta), cos(theta), 0, 0,
    0, 0, 1, 0,
    0, 0, 0, 1]


    Parameters
    ----------
    x : float
        Rotation about the x axis in degrees.
    y : float
        Rotation about the y axis in degrees.
    z : float
        Rotation about the z axis in degrees.

    Returns
    -------
    affine_transform : numpy.ndarray
        A 4x4 affine transformation matrix.
    """
    if x != 0:
        cosine_theta = np.cos(np.deg2rad(x))
        sin_theta = np.sin(np.deg2rad(x))
        x_transform = np.eye(4, 4)
        x_transform[1, 1] = cosine_theta
        x_transform[1, 2] = -sin_theta
        x_transform[2, 1] = sin_theta
        x_transform[2, 2] = cosine_theta
    else:
        x_transform = None

    if y != 0:
        cosine_theta = np.cos(np.deg2rad(y))
        sin_theta = np.sin(np.deg2rad(y))
        y_transform = np.eye(4, 4)
        y_transform[0, 0] = cosine_theta
        y_transform[0, 2] = sin_theta
        y_transform[2, 0] = -sin_theta
        y_transform[2, 2] = cosine_theta
    else:
        y_transform = None

    if z != 0:
        cosine_theta = np.cos(np.deg2rad(z))
        sin_theta = np.sin(np.deg2rad(z))
        z_transform = np.eye(4, 4)
        z_transform[0, 0] = cosine_theta
        z_transform[0, 1] = -sin_theta
        z_transform[1, 0] = sin_theta
        z_transform[1, 1] = cosine_theta
    else:
        z_transform = None

    matrices = [x_transform, y_transform, z_transform]
    matrices = [x for x in matrices if x is not None]

    if len(matrices) == 0:
        return np.eye(4, 4)
    if len(matrices) == 1:
        return matrices[0]
    elif len(matrices) == 2:
        return np.matmul(matrices[0].T, matrices[1])
    else:
        rotation_transform = np.matmul(matrices[0].T, matrices[1])
        return np.matmul(rotation_transform.T, matrices[2])


def affine_shear(dz, dy, dx, dimension="YZ", angle=0):
    """Calculate the general 3D affine transform for shear.

    Returns a shear matrix for a shear about the x, y, and z axes.

    Affine Transform for shear has the following form:

    [1 hxy hxz, 0,
    hyx 1 hyz, 0,
    hzx hzy 1, 0,
    0, 0, 0, 1]

    Parameters
    ----------
    dz : float
        Voxel size in Z.
    dy : float
        Voxel size in Y.
    dx : float
        Voxel size in X.
    dimension : str, optional
        The dimension to shear in. Options are "XY", "XZ", and "YZ".
    angle : float, optional
        The angle to shear in degrees.

    Returns
    -------
    shear_transform : numpy.ndarray
        A 4x4 affine transformation matrix.
    """
    shear_transform = np.eye(4, 4)
    if dz == 0 or dy == 0 or dx == 0 or angle == 0:
        return shear_transform

    scaled_angle = np.multiply(
        np.cos(np.deg2rad(angle)),
        [dy / dx, dz / dx, dz / dy],
    )

    if dimension.upper() == "XY" or dimension.upper() == "YX":
        shear_transform[0, 1] = scaled_angle[0]
    elif dimension.upper() == "XZ" or dimension.upper() == "ZX":
        shear_transform[0, 2] = scaled_angle[1]
    elif dimension.upper() == "YZ" or dimension.upper() == "ZY":
        shear_transform[1, 2] = scaled_angle[2]
    return shear_transform



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging.config
import logging.handlers
from pathlib import Path
import os
import sys
import traceback
from datetime import datetime

# Third Party Imports
import yaml

# Local Imports
from navigate.config.config import get_navigate_path


def update_nested_dict(d, find_func, apply_func):
    """Update a nested dictionary by applying a function to a value

    Loops through a nested dictionary and if find_func() conditions are met,
    run apply_func on that key.

    TODO: This is highly general and doesn't belong here.
    TODO: It might be nice to make this non-recursive.

    Parameters
    ----------
    d : dict
        Dictionary to be updated
    find_func : func
        Accepts key, value pair and matches a condition based on these. Returns bool.
    apply_func : func
        Accepts a value returns the new value.

    Returns
    -------
    d2 : dict
        An version of d, updated according to the passed functions.
    """
    d2 = {}
    for k, v in d.items():
        if find_func(k, v):
            d2[k] = apply_func(v)
        else:
            d2[k] = v
        if isinstance(v, dict):
            d2[k] = update_nested_dict(v, find_func, apply_func)
    return d2


def find_filename(k, v):
    """Check that we've met the condition dictionary key == 'filename'

    Parameters
    ----------
    k : str
        Dictionary key
    v : str
        Dictionary value

    Returns
    -------
    bool
        True if k == 'filename', False otherwise

    Examples
    --------
    >>> find_filename('filename', 'test')
    """
    if k == "filename":
        return True
    return False


def log_setup(logging_configuration, logging_path=None):
    """Setup logging configuration

    Initialize a logger from a YAML file containing information in the Python logging
    dictionary format.

    Note
    ----
        Additional information here:
        https://docs.python.org/3/library/logging.config.html#logging-config-dictschema

    Parameters
    ----------
    logging_configuration : str
        Path to file to be loaded.
        Relative to the location of the folder containing this file.
    logging_path : str, optional
        Path to store logs. Defaults to navigate_path/logs
    """

    # path to logging_configuration is set relative
    # to the location of the folder containing this file (log_functions.py)
    base_directory = Path(__file__).resolve().parent
    logging_configuration_path = Path.joinpath(base_directory, logging_configuration)

    # Save directory for logging information.
    time = datetime.now()
    time_stamp = Path(
        "%s-%s-%s-%s%s"
        % (
            f"{time.year:04d}",
            f"{time.month:02d}",
            f"{time.day:02d}",
            f"{time.hour:02d}",
            f"{time.minute:02d}",
        )
    )

    if logging_path is None:
        logging_path = Path.joinpath(Path(get_navigate_path()), "logs")
    todays_path = Path.joinpath(logging_path, time_stamp)
    if not os.path.exists(logging_path):
        os.mkdir(logging_path)
    if not os.path.exists(todays_path):
        os.mkdir(todays_path)

    def update_filename(v):
        """Function to map filename to base_directory/filename in the dictionary

        Parameters
        ----------
        v : str
            Value to be updated

        Returns
        -------
        Path : str
            Path to the log file
        """
        return Path.joinpath(todays_path, v)

    # Read the logging configuration file.
    with open(logging_configuration_path, "r") as f:
        try:
            config_data = yaml.load(f.read(), Loader=yaml.FullLoader)

            # Force all log files to be created relative to logging_path
            config_data2 = update_nested_dict(
                config_data, find_filename, update_filename
            )
            logging.config.dictConfig(config_data2)

            # Configures our loggers from updated logging.yml
        except yaml.YAMLError as yaml_error:
            print(yaml_error)


def main_process_listener(queue):
    """Listener function for the main process

    This function will listen for new logs put in queue from sub processes,
    it will then log via the main process.

    Parameters
    ----------
    queue : multiprocessing.Queue
        Queue to listen for new logs
    """
    while True:
        try:
            record = queue.get()
            if record is None:
                # Sentinel to tell listener to stop
                break
            logger = logging.getLogger(record.name)
            logger.handle(record)
        except Exception:
            print("Whoops! Problem: ", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
